# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import contextlib
import hashlib
import importlib
import os
import shutil
import tempfile
from hashlib import sha1
from site import makepath  # type: ignore[attr-defined]

from pex import hashing
from pex.common import filter_pyc_dirs, filter_pyc_files, safe_mkdir, safe_mkdtemp
from pex.compatibility import (  # type: ignore[attr-defined]  # `exec_function` is defined dynamically
    PY2,
    exec_function,
)
from pex.orderedset import OrderedSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import IO, Any, Callable, Iterator, Optional

    from pex.hashing import Hasher


class DistributionHelper(object):
    # TODO(#584: This appears unused, but clients might still use it. We cannot remove until we
    #  have a deprecation policy.
    @classmethod
    def access_zipped_assets(cls, static_module_name, static_path, dir_location=None):
        # type: (str, str, Optional[str]) -> str
        """Create a copy of static resource files as we can't serve them from within the pex file.

        :param static_module_name: Module name containing module to cache in a tempdir
        :param static_path: Module name, for example 'serverset'
        :param dir_location: create a new temporary directory inside, or None to have one created
        :returns temp_dir: Temporary directory with the zipped assets inside
        """
        if dir_location is None:
            temp_dir = safe_mkdtemp()
        else:
            temp_dir = dir_location

        module = importlib.import_module(static_module_name)
        # N.B.: This handles namespace packages new and old.
        paths = OrderedSet(os.path.realpath(d) for d in getattr(module, "__path__", []))
        if module.__file__:
            # And this handles old-style __init__.py packages.
            paths.add(os.path.realpath(module.__file__))

        safe_mkdir(temp_dir)
        for path in paths:
            resource_dir = os.path.realpath(os.path.join(path, static_path))
            if os.path.isdir(resource_dir):
                for root, dirs, files in os.walk(resource_dir):
                    for d in dirs:
                        safe_mkdir(
                            os.path.join(
                                temp_dir, os.path.relpath(os.path.join(root, d), resource_dir)
                            )
                        )
                    for f in files:
                        src = os.path.join(root, f)
                        shutil.copy(src, os.path.join(temp_dir, os.path.relpath(src, resource_dir)))
        return temp_dir


class CacheHelper(object):
    @classmethod
    def hash(cls, path, digest=None, hasher=sha1):
        # type: (str, Optional[Hasher], Callable[[], Hasher]) -> str
        """Return the digest of a single file in a memory-efficient manner."""
        if digest is None:
            digest = hasher()
        hashing.file_hash(path, digest)
        return digest.hexdigest()

    @classmethod
    def pex_code_hash(cls, directory):
        # type: (str) -> str
        """Return a reproducible hash of the contents of a loose PEX; excluding all `.pyc` files."""
        digest = hashlib.sha1()
        hashing.dir_hash(
            directory=directory,
            digest=digest,
            dir_filter=filter_pyc_dirs,
            file_filter=lambda files: (f for f in filter_pyc_files(files) if not f.startswith(".")),
        )
        return digest.hexdigest()

    @classmethod
    def dir_hash(cls, directory, digest=None, hasher=sha1):
        # type: (str, Optional[Hasher], Callable[[], Hasher]) -> str
        """Return a reproducible hash of the contents of a directory; excluding all `.pyc` files."""
        if digest is None:
            digest = hasher()
        hashing.dir_hash(
            directory=directory,
            digest=digest,
            dir_filter=filter_pyc_dirs,
            file_filter=filter_pyc_files,
        )
        return digest.hexdigest()

    @classmethod
    def zip_hash(
        cls,
        zip_path,  # type: str
        relpath=None,  # type: Optional[str]
    ):
        # type: (...) -> str
        """Return a reproducible hash of the contents of a zip; excluding all `.pyc` files."""
        digest = hashlib.sha1()
        hashing.zip_hash(
            zip_path=zip_path,
            digest=digest,
            relpath=relpath,
            dir_filter=filter_pyc_dirs,
            file_filter=filter_pyc_files,
        )
        return digest.hexdigest()


@contextlib.contextmanager
def named_temporary_file(**kwargs):
    # type: (**Any) -> Iterator[IO]
    """Due to a bug in python (https://bugs.python.org/issue14243), we need this to be able to use
    the temporary file without deleting it."""
    assert "delete" not in kwargs
    kwargs["delete"] = False
    fp = tempfile.NamedTemporaryFile(**kwargs)
    try:
        with fp:
            yield fp
    finally:
        os.remove(fp.name)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

from threading import Thread

import pytest

from pex.compatibility import PY2
from pex.fetcher import URLFetcher
from pex.typing import TYPE_CHECKING
from pex.version import __version__

if PY2:
    from BaseHTTPServer import BaseHTTPRequestHandler
    from SocketServer import TCPServer
else:
    from http.server import BaseHTTPRequestHandler
    from socketserver import TCPServer

if TYPE_CHECKING:
    from typing import Tuple


@pytest.fixture
def server_address():
    class GETRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            body = self.headers["User-Agent"].encode("utf-8")
            self.send_response(200)
            self.send_header("Content-Type", "application/octet-stream")
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()
            self.wfile.write(body)

    server = TCPServer(("127.0.0.1", 0), GETRequestHandler)
    host, port = server.server_address

    server_thread = Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()
    try:
        yield host, port
    finally:
        server.shutdown()
        server_thread.join()


def test_user_agent(server_address):
    # type: (Tuple[str, int]) -> None

    host, port = server_address
    url = "http://{host}:{port}".format(host=host, port=port)
    url_fetcher = URLFetcher()
    with url_fetcher.get_body_stream(url) as fp:
        assert "pex/{version}".format(version=__version__) == fp.read().decode("utf-8")



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
import platform

from pex.cli.testing import run_pex3
from pex.interpreter import PythonInterpreter
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.targets import CompletePlatform
from pex.testing import IntegResults, built_wheel, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List


def build_pex(
    args,  # type: List[str]
    marker='python_full_version >= "{python_full_version}"'.format(
        python_full_version=platform.python_version()
    ),  # type: str
):
    # type: (...) -> IntegResults

    with built_wheel(
        name="project_has_python_full_version_marker",
        version="0.1.0",
        install_reqs=["ansicolors==1.1.8; {marker}".format(marker=marker)],
        universal=True,
    ) as wheel:
        return run_pex_command(
            args=["-f", os.path.dirname(wheel), "project_has_python_full_version_marker"] + args,
        )


def assert_expected_dists(pex_file):
    # type: (str) -> None
    assert {
        ProjectName("project_has_python_full_version_marker"): Version("0.1.0"),
        ProjectName("ansicolors"): Version("1.1.8"),
    } == {ProjectName(dist.project_name): Version(dist.version) for dist in PEX(pex_file).resolve()}


def test_platform_complete(
    current_interpreter,  # type: PythonInterpreter
    tmpdir,  # type: Any
):
    # type: (...) -> None

    local_interpreter_pex = os.path.join(str(tmpdir), "local_interpreter.pex")
    build_pex(
        args=["-o", local_interpreter_pex, "--python", current_interpreter.binary]
    ).assert_success()
    assert_expected_dists(local_interpreter_pex)

    # An abbreviated platform does not have the Python interpreter patch version needed to form a
    # `python_full_version` marker environment entry.
    result = build_pex(
        args=["-o", local_interpreter_pex, "--platform", "manylinux_2_35_x86_64-cp-310-cp310"]
    )
    result.assert_failure()
    assert "'python_full_version' does not exist in evaluation environment." in result.error

    complete_platform_pex = os.path.join(str(tmpdir), "complete_platform.pex")
    complete_platform = CompletePlatform.from_interpreter(current_interpreter)
    build_pex(
        args=[
            "-o",
            complete_platform_pex,
            "--complete-platform",
            json.dumps(
                dict(
                    marker_environment=complete_platform.marker_environment.as_dict(),
                    compatible_tags=complete_platform.supported_tags.to_string_list(),
                )
            ),
        ]
    ).assert_success()
    assert_expected_dists(complete_platform_pex)

    complete_platform_from_file_pex = os.path.join(str(tmpdir), "complete_platform.from_file.pex")
    platform_file = os.path.join(str(tmpdir), "platform.json")
    run_pex3(
        "interpreter",
        "inspect",
        "--markers",
        "--tags",
        "--python",
        current_interpreter.binary,
        "-o",
        platform_file,
    ).assert_success()
    build_pex(
        args=["-o", complete_platform_from_file_pex, "--complete-platform", platform_file]
    ).assert_success()
    assert_expected_dists(complete_platform_from_file_pex)


def test_platform_abbreviated(
    current_interpreter,  # type: PythonInterpreter
    tmpdir,  # type: Any
):
    # type: (...) -> None

    pex_file = os.path.join(str(tmpdir), "a.pex")

    # An abbreviated platform does not have the Python interpreter patch version needed to form a
    # `python_full_version` marker environment entry.
    result = build_pex(args=["-o", pex_file, "--platform", "manylinux_2_35_x86_64-cp-310-cp310"])
    result.assert_failure()
    assert "'python_full_version' does not exist in evaluation environment." in result.error

    # However, the Platform supplied by a PythonInterpreter (e.g:
    # manylinux_2_35_x86_64-cp-3.10.2-cp310) includes a full version; so it should work.
    current_interpreter_platform = current_interpreter.platform
    assert 3 == len(current_interpreter_platform.version_info)
    build_pex(
        args=["-o", pex_file, "--platform", str(current_interpreter_platform)]
    ).assert_success()
    assert_expected_dists(pex_file)

    # But the Platform supplied by a PythonInterpreter does not keep enough information from the
    # source interpreter to fill in marker environment fields like `platform_version`.
    result = build_pex(
        args=["-o", pex_file, "--platform", str(current_interpreter_platform)],
        marker='platform_version == "#1 SMP PREEMPT Wed, 16 Feb 2022 19:35:18 +0000"',
    )
    result.assert_failure()
    assert "'platform_version' does not exist in evaluation environment." in result.error



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.layout import Layout
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable, List

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class EntryPointArgs(object):
    build_args = attr.ib(default=())  # type: Iterable[str]
    run_args = attr.ib(default=())  # type: Iterable[str]


@pytest.mark.parametrize(
    "execution_mode_args", [pytest.param([], id="PEX"), pytest.param(["--venv"], id="VENV")]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "entry_point_args",
    [
        pytest.param(EntryPointArgs(build_args=["-e", "issues_1018:main"]), id="ep-function"),
        pytest.param(EntryPointArgs(build_args=["-m", "issues_1018"]), id="ep-module"),
        pytest.param(EntryPointArgs(run_args=["-m", "issues_1018"]), id="no-ep"),
    ],
)
def test_execute_module_alter_sys(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
    layout,  # type: Layout.Value
    entry_point_args,  # type: EntryPointArgs
):
    # type: (...) -> None
    src_dir = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src_dir, "issues_1018.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import pickle

                def add(a, b):
                    return a + b

                def main():
                    pickle_add = pickle.dumps(add)
                    add_clone = pickle.loads(pickle_add)
                    print(add_clone(1, 2))

                if __name__ == "__main__":
                    main()
                """
            )
        )
    expected_output = b"3\n"

    # There are 18 ways we can invoke the module above using Pex corresponding to the 3D matrix with
    # axes: execution mode x layout x {entrypoint-function, entrypoint-module, ad-hoc (-m) module}.
    # We support all of these.

    pex_app = os.path.join(str(tmpdir), "app.pex")
    run_pex_command(
        args=["-D", src_dir, "-o", pex_app] + list(entry_point_args.build_args)
    ).assert_success()
    assert expected_output == subprocess.check_output(
        args=[pex_app] + list(entry_point_args.run_args)
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import zipfile
from abc import abstractmethod
from contextlib import contextmanager

from pex.atomic_directory import atomic_directory
from pex.common import is_script, open_zip, safe_copy, safe_mkdir
from pex.enum import Enum
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.variables import unzip_dir

if TYPE_CHECKING:
    from typing import Iterator, Optional

BOOTSTRAP_DIR = ".bootstrap"
DEPS_DIR = ".deps"
PEX_INFO_PATH = "PEX-INFO"
PEX_LAYOUT_PATH = "PEX-LAYOUT"


class Layout(Enum["Layout.Value"]):
    class Value(Enum.Value):
        @classmethod
        def try_load(cls, pex_directory):
            # type: (str) -> Optional[Layout.Value]
            layout = os.path.join(pex_directory, PEX_LAYOUT_PATH)
            if not os.path.isfile(layout):
                return None
            with open(layout) as fp:
                return Layout.for_value(fp.read().strip())

        def record(self, pex_directory):
            # type: (str) -> None
            with open(os.path.join(pex_directory, PEX_LAYOUT_PATH), "w") as fp:
                fp.write(self.value)

    ZIPAPP = Value("zipapp")
    PACKED = Value("packed")
    LOOSE = Value("loose")

    @classmethod
    def identify(cls, pex):
        # type: (str) -> Layout.Value
        """Assumes pex is a valid PEX and identifies its layout."""
        if zipfile.is_zipfile(pex) and is_script(
            pex,
            # N.B.: A PEX file need not be executable since it can always be run via `python a.pex`.
            check_executable=False,
        ):
            return cls.ZIPAPP

        if os.path.isdir(pex) and zipfile.is_zipfile(os.path.join(pex, BOOTSTRAP_DIR)):
            return cls.PACKED

        return cls.LOOSE

    @classmethod
    def identify_original(cls, pex):
        # type: (str) -> Layout.Value
        layout = cls.identify(pex)
        if layout != Layout.LOOSE:
            return layout
        return cls.Value.try_load(pex) or Layout.LOOSE


class _Layout(object):
    def __init__(
        self,
        layout,  # type: Layout.Value
        path,  # type: str
    ):
        # type: (...) -> None
        self._layout = layout
        self._path = os.path.normpath(path)

    @property
    def path(self):
        # type: () -> str
        return self._path

    def bootstrap_strip_prefix(self):
        # type: () -> Optional[str]
        return None

    @abstractmethod
    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    def dist_strip_prefix(self, dist_name):
        # type: (str) -> Optional[str]
        return None

    @abstractmethod
    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
    ):
        # type: (...) -> None
        raise NotImplementedError()

    @abstractmethod
    def extract_code(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def extract_main(self, dest_dir):
        # type: (str) -> None
        raise NotImplementedError()

    def record(self, dest_dir):
        # type: (str) -> None
        self._layout.record(dest_dir)


def _install(
    layout,  # type: _Layout
    pex_root,  # type: str
    pex_hash,  # type: str
):
    # type: (...) -> str
    with TRACER.timed("Laying out {}".format(layout)):
        pex = layout.path
        install_to = unzip_dir(pex_root=pex_root, pex_hash=pex_hash)
        with atomic_directory(install_to) as chroot:
            if not chroot.is_finalized():
                with TRACER.timed("Installing {} to {}".format(pex, install_to)):
                    from pex.pex_info import PexInfo

                    pex_info = PexInfo.from_pex(pex)
                    pex_info.update(PexInfo.from_env())

                    bootstrap_cache = pex_info.bootstrap_cache
                    if bootstrap_cache is None:
                        raise AssertionError(
                            "Expected bootstrap_cache to be populated for {}.".format(layout)
                        )
                    code_hash = pex_info.code_hash
                    if code_hash is None:
                        raise AssertionError(
                            "Expected code_hash to be populated for {}.".format(layout)
                        )

                    with atomic_directory(
                        bootstrap_cache, source=layout.bootstrap_strip_prefix()
                    ) as bootstrap_zip_chroot:
                        if not bootstrap_zip_chroot.is_finalized():
                            layout.extract_bootstrap(bootstrap_zip_chroot.work_dir)
                    os.symlink(
                        os.path.join(os.path.relpath(bootstrap_cache, install_to)),
                        os.path.join(chroot.work_dir, BOOTSTRAP_DIR),
                    )

                    for location, sha in pex_info.distributions.items():
                        spread_dest = os.path.join(pex_info.install_cache, sha, location)
                        dist_relpath = os.path.join(DEPS_DIR, location)
                        with atomic_directory(
                            spread_dest,
                            source=layout.dist_strip_prefix(location),
                        ) as spread_chroot:
                            if not spread_chroot.is_finalized():
                                layout.extract_dist(spread_chroot.work_dir, dist_relpath)
                        symlink_dest = os.path.join(chroot.work_dir, dist_relpath)
                        safe_mkdir(os.path.dirname(symlink_dest))
                        os.symlink(
                            os.path.relpath(
                                spread_dest,
                                os.path.join(install_to, os.path.dirname(dist_relpath)),
                            ),
                            symlink_dest,
                        )

                    code_dest = os.path.join(pex_info.zip_unsafe_cache, code_hash)
                    with atomic_directory(code_dest) as code_chroot:
                        if not code_chroot.is_finalized():
                            layout.extract_code(code_chroot.work_dir)
                    for path in os.listdir(code_dest):
                        os.symlink(
                            os.path.join(os.path.relpath(code_dest, install_to), path),
                            os.path.join(chroot.work_dir, path),
                        )

                    layout.extract_pex_info(chroot.work_dir)
                    layout.extract_main(chroot.work_dir)
                    layout.record(chroot.work_dir)
        return install_to


class _ZipAppPEX(_Layout):
    def __init__(
        self,
        path,  # type: str
        zfp,  # type: zipfile.ZipFile
    ):
        # type: (...) -> None
        super(_ZipAppPEX, self).__init__(Layout.ZIPAPP, path)
        self._zfp = zfp
        self._names = tuple(zfp.namelist())

    def bootstrap_strip_prefix(self):
        # type: () -> Optional[str]
        return BOOTSTRAP_DIR

    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        for name in self._names:
            if name.startswith(BOOTSTRAP_DIR) and not name.endswith("/"):
                self._zfp.extract(name, dest_dir)

    def dist_strip_prefix(self, dist_name):
        # type: (str) -> Optional[str]
        return os.path.join(DEPS_DIR, dist_name)

    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
    ):
        for name in self._names:
            if name.startswith(dist_relpath) and not name.endswith("/"):
                self._zfp.extract(name, dest_dir)

    def extract_code(self, dest_dir):
        # type: (str) -> None
        for name in self._names:
            if name not in ("__main__.py", PEX_INFO_PATH) and not name.startswith(
                (BOOTSTRAP_DIR, DEPS_DIR)
            ):
                self._zfp.extract(name, dest_dir)

    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        self._zfp.extract(PEX_INFO_PATH, dest_dir)

    def extract_main(self, dest_dir):
        # type: (str) -> None
        self._zfp.extract("__main__.py", dest_dir)

    def __str__(self):
        return "PEX zipfile {}".format(self._path)


class _PackedPEX(_Layout):
    def __init__(self, path):
        # type: (str) -> None
        super(_PackedPEX, self).__init__(Layout.PACKED, path)

    def extract_bootstrap(self, dest_dir):
        # type: (str) -> None
        with open_zip(os.path.join(self._path, BOOTSTRAP_DIR)) as zfp:
            zfp.extractall(dest_dir)

    def extract_dist(
        self,
        dest_dir,  # type: str
        dist_relpath,  # type: str
    ):
        with open_zip(os.path.join(self._path, dist_relpath)) as zfp:
            zfp.extractall(dest_dir)

    def extract_code(self, dest_dir):
        # type: (str) -> None
        for root, dirs, files in os.walk(self._path):
            rel_root = os.path.relpath(root, self._path)
            if root == self._path:
                dirs[:] = [d for d in dirs if d != DEPS_DIR]
                files[:] = [
                    f for f in files if f not in ("__main__.py", PEX_INFO_PATH, BOOTSTRAP_DIR)
                ]
            for d in dirs:
                safe_mkdir(os.path.join(dest_dir, rel_root, d))
            for f in files:
                safe_copy(
                    os.path.join(root, f),
                    os.path.join(dest_dir, rel_root, f),
                )

    def extract_pex_info(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, PEX_INFO_PATH), os.path.join(dest_dir, PEX_INFO_PATH))

    def extract_main(self, dest_dir):
        # type: (str) -> None
        safe_copy(os.path.join(self._path, "__main__.py"), os.path.join(dest_dir, "__main__.py"))

    def __str__(self):
        return "Spread PEX directory {}".format(self._path)


@contextmanager
def _identify_layout(pex):
    # type: (str) -> Iterator[Optional[_Layout]]

    layout = Layout.identify(pex)
    if Layout.ZIPAPP is layout:
        with open_zip(pex) as zfp:
            yield _ZipAppPEX(pex, zfp)
    elif Layout.PACKED is layout:
        yield _PackedPEX(pex)
    else:
        # A loose PEX which needs no layout.
        yield None


def maybe_install(
    pex,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
):
    # type: (...) -> Optional[str]
    """Installs a zipapp or packed PEX into the pex root as a loose PEX.

    Returns the path of the installed PEX or `None` if the PEX needed no installation and can be
    executed directly.
    """
    with _identify_layout(pex) as layout:
        if layout:
            return _install(layout, pex_root, pex_hash)
    return None



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

import pytest

from pex.dist_metadata import CallableEntryPoint, DistMetadata, Distribution, EntryPoint
from pex.finders import get_entry_point_from_console_script, get_script_from_distributions
from pex.pep_376 import InstalledWheel
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.installation import get_pip
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict

    import attr  # vendor:skip
else:
    from pex.third_party import attr


# In-part, tests a bug where the wheel distribution name has dashes as reported in:
#   https://github.com/pantsbuild/pex/issues/443
#   https://github.com/pantsbuild/pex/issues/551
def test_get_script_from_distributions(tmpdir):
    # type: (Any) -> None
    whl_path = "./tests/example_packages/aws_cfn_bootstrap-1.4-py2-none-any.whl"
    install_dir = os.path.join(str(tmpdir), os.path.basename(whl_path))
    get_pip().spawn_install_wheel(wheel=whl_path, install_dir=install_dir).wait()

    dist = Distribution.load(install_dir)
    assert "aws-cfn-bootstrap" == dist.project_name

    dist_script = get_script_from_distributions("cfn-signal", [dist])
    assert dist_script is not None
    assert dist_script.dist is dist
    assert InstalledWheel.load(install_dir).stashed_path("bin/cfn-signal") == dist_script.path
    assert dist_script.read_contents().startswith(
        b"#!"
    ), "Expected a `scripts`-style script w/shebang."

    assert None is get_script_from_distributions("non_existent_script", [dist])


def create_dist(
    key,  # str
    console_script_entry,  # type: str
):
    # type: (...) -> Distribution
    entry_point = EntryPoint.parse(console_script_entry)

    @attr.s(frozen=True)
    class FakeDist(Distribution):
        def get_entry_map(self):
            # type: () -> Dict[str, Dict[str, EntryPoint]]
            return {"console_scripts": {entry_point.name: entry_point}}

    return FakeDist(
        location=os.getcwd(),
        metadata=DistMetadata(
            project_name=ProjectName(key),
            version=Version("1.0"),
        ),
    )


def test_get_entry_point_from_console_script():
    # type: () -> None
    dists = [
        create_dist(key="fake", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]

    dist_entrypoint = get_entry_point_from_console_script("bob", dists)
    assert dist_entrypoint is not None
    assert (
        CallableEntryPoint(name="bob", module="bob.main", attrs=("run",))
        == dist_entrypoint.entry_point
    )
    assert dist_entrypoint.dist in dists


def test_get_entry_point_from_console_script_conflict():
    # type: () -> None
    dists = [
        create_dist(key="bob", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]
    with pytest.raises(RuntimeError):
        get_entry_point_from_console_script("bob", dists)


def test_get_entry_point_from_console_script_dne():
    # type: () -> None
    dists = [
        create_dist(key="bob", console_script_entry="bob= bob.main:run"),
        create_dist(key="fake", console_script_entry="bob =bob.main:run"),
    ]
    assert get_entry_point_from_console_script("jane", dists) is None



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import re

from pex import hashing
from pex.common import filter_pyc_dirs, filter_pyc_files, open_zip, temporary_dir
from pex.hashing import Sha256
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.requirements import VCS
from pex.resolve.resolved_requirement import Fingerprint
from pex.result import Error, try_
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple, Union

    from pex.hashing import HintedDigest


def _find_built_source_dist(
    build_dir,  # type: str
    project_name,  # type: ProjectName
    version,  # type: Version
):
    # type: (...) -> Union[str, Error]

    # All VCS requirements are prepared as zip archives with this naming scheme as
    # encoded in: `pip._internal.req.req_install.InstallRequirement.archive`.

    listing = os.listdir(build_dir)
    pattern = re.compile(
        r"{project_name}-(?P<version>.+)\.zip".format(
            project_name=project_name.normalized.replace("-", "[-_.]+")
        )
    )
    for name in listing:
        match = pattern.match(name)
        if match and Version(match.group("version")) == version:
            return os.path.join(build_dir, name)

    return Error(
        "Expected to find built sdist for {project_name} {version} in {build_dir} but only found:\n"
        "{listing}".format(
            project_name=project_name.raw,
            version=version.raw,
            build_dir=build_dir,
            listing="\n".join(listing),
        )
    )


def fingerprint_downloaded_vcs_archive(
    download_dir,  # type: str
    project_name,  # type: str
    version,  # type: str
    vcs,  # type: VCS.Value
):
    # type: (...) -> Tuple[Fingerprint, str]

    archive_path = try_(
        _find_built_source_dist(
            build_dir=download_dir, project_name=ProjectName(project_name), version=Version(version)
        )
    )
    digest = Sha256()
    digest_vcs_archive(archive_path=archive_path, vcs=vcs, digest=digest)
    return Fingerprint.from_digest(digest), archive_path


def digest_vcs_archive(
    archive_path,  # type: str
    vcs,  # type: VCS.Value
    digest,  # type: HintedDigest
):
    # type: (...) -> None

    # All VCS requirements are prepared as zip archives as encoded in:
    # `pip._internal.req.req_install.InstallRequirement.archive`.
    with TRACER.timed(
        "Digesting {archive} {vcs} archive".format(archive=os.path.basename(archive_path), vcs=vcs)
    ), temporary_dir() as chroot, open_zip(archive_path) as archive:
        archive.extractall(chroot)

        # Ignore VCS control directories for the purposes of fingerprinting the version controlled
        # source tree. VCS control directories can contain non-reproducible content (Git at least
        # has files that contain timestamps).
        #
        # We cannot prune these directories from the source archive directly unfortunately since
        # some build processes use VCS version information to derive their version numbers (C.F.:
        # https://pypi.org/project/setuptools-scm/). As such, we'll get a stable fingerprint, but be
        # forced to re-build a wheel each time the VCS requirement is re-locked later, even when it
        # hashes the same.
        vcs_control_dir = ".{vcs}".format(vcs=vcs)

        # TODO(John Sirois): Consider implementing zip_hash to avoid the extractall.
        hashing.dir_hash(
            directory=chroot,
            digest=digest,
            dir_filter=lambda dirs: [d for d in filter_pyc_dirs(dirs) if d != vcs_control_dir],
            file_filter=filter_pyc_files,
        )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import shutil
import subprocess
from textwrap import dedent

import pytest

from pex.testing import IS_LINUX, IS_MAC, PY_VER, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    PY_VER < (3, 6), reason="The mypy_protobuf 2.4 distribution is only available for Python 3.6+"
)
@pytest.mark.xfail(
    IS_MAC,
    reason=(
        "On modern Linux (starting with the 5.1 kernel shipped on May 19th 2019), the default max "
        "shebang length limit is 256 but the hardcoded limit in Pip that #1520 fixes is 127; so "
        "the work-around here should test green on Linux. On Mac, however, the hardcoded limit in "
        "Pip that #1520 fixes is 512 and that limit has been stable on macOS; so we expect the PEX "
        "creation to fail with something like: [Errno 63] File name too long: '/tmp/"
        "pytest-of-runner/pytest-0/popen-gw2/test_hermetic_console_scripts0/<512 of `_`>/pex_root/"
        "isolated/.488310d43ea7ca80b559c306f2db44914a184e37.atomic_directory.lck'."
    ),
)
def test_hermetic_console_scripts(tmpdir):
    # type: (Any) -> None

    # N.B.: See pex/vendor/_vendored/pip/pip/_vendor/distlib/scripts.py lines 127-156.
    # https://github.com/pantsbuild/pex/blob/196b4cd5b8dd4b4af2586460530e9a777262be7d/pex/vendor/_vendored/pip/pip/_vendor/distlib/scripts.py#L127-L156
    length_pad = 127 if IS_LINUX else 512
    pex_root = os.path.join(str(tmpdir), "_" * length_pad, "pex_root")
    assert len(pex_root) > length_pad

    mypy_protobuf_pex = os.path.join(str(tmpdir), "mypy_protobuf.pex")

    # Although mypy_protobuf 2.4 is quite old, it depends on an open ended-protobuf and is broken
    # in conjunction with protobuf 4+
    constraints = os.path.join(str(tmpdir), "constraints.txt")
    with open(constraints, "w") as fp:
        fp.write("protobuf<4")

    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "mypy_protobuf==2.4",
            "--constraints",
            constraints,
            "-o",
            mypy_protobuf_pex,
            "--venv",
            "prepend",
        ],
    ).assert_success()

    scripts = [
        os.path.join(root, f)
        for root, dirs, files in os.walk(os.path.join(pex_root, "installed_wheels"))
        for f in files
        if "protoc-gen-mypy" == f
    ]
    assert 1 == len(scripts)
    with open(scripts[0]) as fp:
        assert "#!python" == fp.readline().strip()
        assert "# -*- coding: utf-8 -*-" == fp.readline().strip()

    shutil.rmtree(pex_root)
    # This should no-op (since there is no proto sent on stdin) and exit success.
    subprocess.check_call(
        args=[
            mypy_protobuf_pex,
            "-c",
            dedent(
                """\
                import subprocess
                import sys


                process = subprocess.Popen(['protoc-gen-mypy'], stdin=subprocess.PIPE)
                process.communicate()
                sys.exit(process.returncode)
                """
            ),
        ],
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import re
import subprocess

from pex.cli.testing import run_pex3
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraint
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_lock_create_sdist_requires_python_different_from_current(
    tmpdir,  # type: Any
    py27,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock")
    create_lock_args = [
        "lock",
        "create",
        "--resolver-version",
        "pip-2020-resolver",
        "--style",
        "universal",
        "--interpreter-constraint",
        "CPython<3.11,>=3.8",
        "--python-path",
        os.pathsep.join(interp.binary for interp in (py27, py38, py310)),
        "aioconsole==0.4.1",
        "-o",
        lock,
        "--indent",
        "2",
    ]

    # 1st prove this does the wrong thing on prior broken versions of Pex.
    result = run_pex_command(
        args=["pex==2.1.82", "-c", "pex3", "--"] + create_lock_args,
        python=py27.binary,
        quiet=True,
    )
    result.assert_failure()
    assert (
        "ERROR: Package 'aioconsole' requires a different Python: {pyver} not in '>=3.7'".format(
            pyver=py27.identity.version_str
        )
        in result.error.splitlines()
    )

    # Now show it currently works.
    subprocess.check_call(
        args=[py27.binary, "-m", "pex.cli"] + create_lock_args + ["--pip-version", "20.3.4-patched"]
    )
    run_pex_command(
        args=["--lock", lock, "--", "-c", "import aioconsole"],
        python=py310.binary,
    ).assert_success()


def test_lock_create_universal_interpreter_constraint_unsatisfiable(
    tmpdir,  # type: Any
    py27,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock")
    result = run_pex3(
        "lock",
        "create",
        "--resolver-version",
        "pip-2020-resolver",
        "--style",
        "universal",
        "--interpreter-constraint",
        "CPython<3.11,>=3.9",
        "--python-path",
        os.pathsep.join(interp.binary for interp in (py27, py38)),
        "aioconsole==0.4.1",
        "-o",
        lock,
        "--indent",
        "2",
    )
    result.assert_failure()
    assert re.match(
        r"^When creating a universal lock with an --interpreter-constraint, an interpreter "
        r"matching the constraint must be found on the local system but none was: Could not find a "
        r"compatible interpreter\.\n"
        r"\n"
        r"Examined the following interpreters:\n"
        r"1\.\)\s+{py27_path} {py27_req}\n"
        r"2\.\)\s+{py38_path} {py38_req}\n"
        r"\n"
        r"No interpreter compatible with the requested constraints was found:\n"
        r"\n"
        r"  Version matches CPython<3\.11,>=3\.9\n".format(
            py27_path=py27.binary,
            py27_req=InterpreterConstraint.exact_version(py27),
            py38_path=py38.binary,
            py38_req=InterpreterConstraint.exact_version(py38),
        ),
        result.error,
    )



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys
from contextlib import contextmanager

from pex import third_party
from pex.common import temporary_dir
from pex.compatibility import commonpath
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Dict, Iterator, Tuple


@contextmanager
def temporary_pex_root():
    # type: () -> Iterator[Tuple[str, Dict[str, str]]]
    with temporary_dir() as pex_root, ENV.patch(PEX_ROOT=os.path.realpath(pex_root)) as env:
        original_isolated = third_party._ISOLATED
        try:
            third_party._ISOLATED = None
            yield os.path.realpath(pex_root), env
        finally:
            third_party._ISOLATED = original_isolated


def test_isolated_pex_root():
    # type: () -> None
    with temporary_pex_root() as (pex_root, _):
        devendored_chroot = os.path.realpath(third_party.isolated().chroot_path)
        assert pex_root == commonpath([pex_root, devendored_chroot])


def test_isolated_vendoring_constraints_omitted():
    # type: () -> None
    with temporary_pex_root() as (pex_root, _):
        devendored_chroot = os.path.realpath(third_party.isolated().chroot_path)
        assert [] == [
            os.path.join(root, file)
            for root, _, files in os.walk(devendored_chroot)
            for file in files
            if file == "constraints.txt"
        ]


def test_isolated_idempotent_inprocess():
    # type: () -> None
    with temporary_pex_root():
        result1 = third_party.isolated()
        result2 = third_party.isolated()
        assert result1.pex_hash == result2.pex_hash
        assert os.path.realpath(result1.chroot_path) == os.path.realpath(result2.chroot_path)


def test_isolated_idempotent_subprocess():
    # type: () -> None
    with temporary_pex_root() as (_, env):
        devendored_chroot = os.path.realpath(third_party.isolated().chroot_path)
        stdout = subprocess.check_output(
            args=[
                sys.executable,
                "-c",
                "from pex import third_party; print(third_party.isolated().chroot_path)",
            ],
            env=env,
        )
        assert devendored_chroot == os.path.realpath(stdout.decode("utf-8").strip())



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from textwrap import dedent

import pytest

from pex import targets
from pex.dist_metadata import Requirement
from pex.interpreter import PythonInterpreter
from pex.pep_425 import CompatibilityTags, TagRank
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.requirements import VCS
from pex.resolve.locked_resolve import (
    Artifact,
    DownloadableArtifact,
    FileArtifact,
    LocalProjectArtifact,
    LockedRequirement,
    LockedResolve,
    RankedArtifact,
    Resolved,
    VCSArtifact,
    _ResolvedArtifact,
)
from pex.resolve.resolved_requirement import Fingerprint, PartialArtifact, Pin
from pex.result import Error, try_
from pex.sorted_tuple import SortedTuple
from pex.targets import AbbreviatedPlatform, CompletePlatform, LocalInterpreter, Target
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.third_party.packaging.tags import Tag
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Union


@pytest.fixture
def current_target():
    # type: () -> Target
    return targets.current()


@pytest.fixture
def py37_target(py38):
    # type: (PythonInterpreter) -> Target
    return LocalInterpreter.create(py38)


@pytest.fixture
def py310_target(py310):
    # type: (PythonInterpreter) -> Target
    return LocalInterpreter.create(py310)


def pin(
    project_name,  # type: str
    version,  # type: str
):
    # type: (...) -> Pin
    return Pin(project_name=ProjectName(project_name), version=Version(version))


def artifact(
    url,  # type: str
    algorithm,  # type: str
    hash,  # type: str
):
    # type: (...) -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    return Artifact.from_url(url=url, fingerprint=Fingerprint(algorithm=algorithm, hash=hash))


def locked_requirements(*locked_requirements):
    # type: (*LockedRequirement) -> SortedTuple[LockedRequirement]
    return SortedTuple(locked_requirements)


def req(requirement):
    # type: (str) -> Requirement
    return Requirement.parse(requirement)


def requirements(*reqs):
    # type: (*str) -> Iterable[Requirement]
    return tuple(req(requirement) for requirement in reqs)


@pytest.fixture
def ansicolors_simple():
    # type: () -> LockedResolve
    return LockedResolve(
        platform_tag=Tag("cp27", "cp27mu", "linux_x86_64"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.7"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                    algorithm="blake256",
                    hash="cafebabe",
                ),
                additional_artifacts=[
                    artifact(
                        url="https://example.org/ansicolors-1.1.7.tar.gz",
                        algorithm="sha256",
                        hash="cafebabe",
                    ),
                ],
            ),
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.8"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.8.tar.gz",
                    algorithm="md5",
                    hash="cafebabe",
                ),
            ),
        ),
    )


def assert_resolved(
    result,  # type: Union[Resolved, Error]
    *downloadable_artifacts  # type: DownloadableArtifact
):
    # type: (...) -> None
    assert SortedTuple(downloadable_artifacts) == try_(result).downloadable_artifacts


def test_build(
    current_target,  # type: Target
    ansicolors_simple,  # type: LockedResolve
):
    # type: (...) -> None
    assert_resolved(
        ansicolors_simple.resolve(current_target, [req("ansicolors")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8.tar.gz",
                algorithm="md5",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )
    assert_resolved(
        ansicolors_simple.resolve(current_target, [req("ansicolors")], build=False),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )


def test_use_wheel(
    current_target,  # type: Target
    ansicolors_simple,  # type: LockedResolve
):
    # type: (...) -> None
    assert_resolved(
        ansicolors_simple.resolve(current_target, [req("ansicolors==1.1.7")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors==1.1.7"),
        ),
    )
    assert_resolved(
        ansicolors_simple.resolve(current_target, [req("ansicolors==1.1.7")], use_wheel=False),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7.tar.gz",
                algorithm="sha256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors==1.1.7"),
        ),
    )


def test_constraints(
    current_target,  # type: Target
    ansicolors_simple,  # type: LockedResolve
):
    # type: (...) -> None
    assert_resolved(
        ansicolors_simple.resolve(current_target, [req("ansicolors")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8.tar.gz",
                algorithm="md5",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )
    assert_resolved(
        ansicolors_simple.resolve(
            current_target, [req("ansicolors")], constraints=[req("ansicolors<1.1.8")]
        ),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )


@pytest.fixture
def ansicolors_exotic():
    # type: () -> LockedResolve
    return LockedResolve(
        platform_tag=Tag("cp37", "cp37m", "exotic"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.8"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.8-cp37-cp37m-exotic.whl",
                    algorithm="blake256",
                    hash="cafebabe",
                ),
                additional_artifacts=[
                    artifact(
                        url="https://example.org/ansicolors-1.1.8-cp37-abi3-exotic.whl",
                        algorithm="sha256",
                        hash="cafebabe",
                    ),
                    artifact(
                        url="https://example.org/ansicolors-1.1.8.tar.gz",
                        algorithm="md5",
                        hash="cafebabe",
                    ),
                ],
            ),
        ),
    )


def assert_error(
    result,  # type: Union[Resolved, Error]
    expected_error_message,  # type: str
):
    assert Error(expected_error_message.strip()) == result


def platform(plat):
    # type: (str) -> AbbreviatedPlatform
    return AbbreviatedPlatform.create(Platform.create(plat))


def test_invalid_configuration(
    current_target,  # type: Target
    ansicolors_exotic,  # type: LockedResolve
):
    # type: (...) -> None
    assert_error(
        ansicolors_exotic.resolve(
            current_target, [req("ansicolors")], build=False, use_wheel=False
        ),
        "Cannot both ignore wheels (use_wheel=False) and refrain from building distributions "
        "(build=False).",
    )


def test_platform_resolve(ansicolors_exotic):
    # type: (LockedResolve) -> None

    assert_resolved(
        ansicolors_exotic.resolve(platform("exotic-cp-37-m"), [req("ansicolors")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8-cp37-cp37m-exotic.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )

    assert_resolved(
        ansicolors_exotic.resolve(platform("exotic-cp-38-cp38"), [req("ansicolors")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8-cp37-abi3-exotic.whl",
                algorithm="sha256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )


def test_not_found(
    current_target,  # type: Target
    ansicolors_exotic,  # type: LockedResolve
):
    # type: (...) -> None
    assert_error(
        ansicolors_exotic.resolve(current_target, [req("requests>1")]),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: True

            Dependency on requests (via: requests>1) not satisfied, no candidates found.
            """
        ).format(target_description=current_target.render_description()),
    )


def test_source(
    current_target,  # type: Target
    ansicolors_exotic,  # type: LockedResolve
):
    # type: (...) -> None
    assert_error(
        ansicolors_exotic.resolve(current_target, [req("requests>1")], source="lock.json"),
        dedent(
            """\
            Failed to resolve all requirements for {target_description} from lock.json:

            Configured with:
                build: True
                use_wheel: True

            Dependency on requests (via: requests>1) not satisfied, no candidates found.
            """
        ).format(target_description=current_target.render_description()),
    )


def test_version_mismatch(
    current_target,  # type: Target
    ansicolors_exotic,  # type: LockedResolve
    ansicolors_simple,  # type: LockedResolve
):
    # type: (...) -> None

    assert_error(
        ansicolors_exotic.resolve(current_target, [req("ansicolors>1,<1.1")]),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: True

            Dependency on ansicolors not satisfied, 1 incompatible candidate found:
            1.) ansicolors 1.1.8 does not satisfy the following requirements:
                <1.1,>1 (via: ansicolors<1.1,>1)
            """
        ).format(target_description=current_target.render_description()),
    )

    assert_error(
        ansicolors_simple.resolve(current_target, [req("ansicolors>1,<1.1")]),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: True

            Dependency on ansicolors not satisfied, 2 incompatible candidates found:
            1.) ansicolors 1.1.7 does not satisfy the following requirements:
                <1.1,>1 (via: ansicolors<1.1,>1)
            2.) ansicolors 1.1.8 does not satisfy the following requirements:
                <1.1,>1 (via: ansicolors<1.1,>1)
            """
        ).format(target_description=current_target.render_description()),
    )


def test_wheel_tag_mismatch(
    current_target,  # type: Target
    ansicolors_exotic,  # type: LockedResolve
):
    # type: (...) -> None
    assert_error(
        ansicolors_exotic.resolve(current_target, [req("ansicolors==1.1.*")], build=False),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: False
                use_wheel: True

            Dependency on ansicolors not satisfied, 1 incompatible candidate found:
            1.) ansicolors 1.1.8 (via: ansicolors==1.1.*) does not have any compatible artifacts:
                https://example.org/ansicolors-1.1.8-cp37-cp37m-exotic.whl
                https://example.org/ansicolors-1.1.8-cp37-abi3-exotic.whl
                https://example.org/ansicolors-1.1.8.tar.gz
            """
        ).format(target_description=current_target.render_description()),
    )

    assert_resolved(
        ansicolors_exotic.resolve(current_target, [req("ansicolors==1.1.*")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8.tar.gz",
                algorithm="md5",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors==1.1.*"),
        ),
    )


def test_requires_python_mismatch(
    py37_target,  # type: Target
    py310_target,  # type: Target
):
    # type: (...) -> None

    locked_resolve = LockedResolve(
        platform_tag=Tag("cp37", "cp37m", "exotic"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.7"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                    algorithm="blake256",
                    hash="cafebabe",
                ),
                requires_python=SpecifierSet(">=3,<3.10"),
            ),
        ),
    )

    assert_error(
        locked_resolve.resolve(py310_target, [req("ansicolors==1.1.7")]),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: True

            Dependency on ansicolors not satisfied, 1 incompatible candidate found:
            1.) ansicolors 1.1.7 (via: ansicolors==1.1.7) requires Python <3.10,>=3
            """
        ).format(target_description=py310_target.render_description()),
    )

    assert_resolved(
        locked_resolve.resolve(py37_target, [req("ansicolors==1.1.7")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors==1.1.7"),
        ),
    )


def test_constraint_mismatch(
    current_target,  # type: Target
    ansicolors_simple,  # type: LockedResolve
):
    # type: (...) -> None
    locked_resolve = LockedResolve(
        platform_tag=Tag("cp37", "cp37m", "exotic"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.7"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                    algorithm="blake256",
                    hash="cafebabe",
                ),
            ),
        ),
    )

    assert_error(
        locked_resolve.resolve(
            current_target, [req("ansicolors")], constraints=[req("ansicolors>=2")]
        ),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: True

            Dependency on ansicolors not satisfied, 1 incompatible candidate found:
            1.) ansicolors 1.1.7 does not satisfy the following requirements:
                >=2 (via: constraint)
            """
        ).format(target_description=current_target.render_description()),
    )

    assert_resolved(
        locked_resolve.resolve(
            current_target, [req("ansicolors")], constraints=[req("irrelevant==1.0.0")]
        ),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors"),
        ),
    )


def test_prefer_older_binary(current_target):
    # type: (Target) -> None

    locked_resolve = LockedResolve(
        platform_tag=Tag("cp37", "cp37m", "exotic"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.7"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                    algorithm="blake256",
                    hash="cafebabe",
                ),
            ),
            LockedRequirement.create(
                pin=pin("ansicolors", "1.1.8"),
                artifact=artifact(
                    url="https://example.org/ansicolors-1.1.8.tar.gz",
                    algorithm="sha256",
                    hash="cafebabe",
                ),
            ),
        ),
    )

    assert_resolved(
        locked_resolve.resolve(current_target, [req("ansicolors>1")]),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.8"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.8.tar.gz",
                algorithm="sha256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors>1"),
        ),
    )

    assert_resolved(
        locked_resolve.resolve(current_target, [req("ansicolors>1")], prefer_older_binary=True),
        DownloadableArtifact.create(
            pin=pin("ansicolors", "1.1.7"),
            artifact=artifact(
                url="https://example.org/ansicolors-1.1.7-1buildtag-py2.py3-none-any.whl",
                algorithm="blake256",
                hash="cafebabe",
            ),
            satisfied_direct_requirements=requirements("ansicolors>1"),
        ),
    )


def resolved_artifact(
    project_name,  # type: str
    version,  # type: str
    artifact_basename,  # type: str
    rank_value,  # type: int
):
    # type: (...) -> _ResolvedArtifact
    primary_artifact = artifact(
        url="https://example.org/{artifact_basename}".format(artifact_basename=artifact_basename),
        algorithm="sha256",
        hash="cafebabe",
    )
    return _ResolvedArtifact(
        ranked_artifact=RankedArtifact(artifact=primary_artifact, rank=TagRank(rank_value)),
        locked_requirement=LockedRequirement.create(
            pin=pin(project_name, version), artifact=primary_artifact
        ),
    )


def test_resolved_artifact_select_higher_rank():
    specific_wheel_117 = resolved_artifact(
        "ansicolors", "1.1.7", "ansicolors-1.1.7-cp38-cp38-linux_x86_64.whl", 0
    )
    specific_wheel_118 = resolved_artifact(
        "ansicolors", "1.1.8", "ansicolors-1.1.8-cp38-cp38-linux_x86_64.whl", 0
    )

    general_wheel_117 = resolved_artifact(
        "ansicolors", "1.1.7", "ansicolors-1.1.7-py3-none-any.whl", 122
    )
    general_wheel_118 = resolved_artifact(
        "ansicolors", "1.1.8", "ansicolors-1.1.8-py3-none-any.whl", 122
    )

    source_117 = resolved_artifact("ansicolors", "1.1.7", "ansicolors-1.1.7.tar.gz", 123)
    source_118 = resolved_artifact("ansicolors", "1.1.8", "ansicolors-1.1.8.tar.gz", 123)

    # Same rank, different versions.
    assert specific_wheel_118 is specific_wheel_117.select_higher_rank(specific_wheel_118)
    assert specific_wheel_118 is specific_wheel_117.select_higher_rank(
        specific_wheel_118, prefer_older_binary=True
    )
    assert general_wheel_118 is general_wheel_118.select_higher_rank(general_wheel_117)
    assert general_wheel_118 is general_wheel_118.select_higher_rank(
        general_wheel_117, prefer_older_binary=True
    )
    assert source_118 is source_117.select_higher_rank(source_118)
    assert source_118 is source_117.select_higher_rank(source_118, prefer_older_binary=True)

    # Same version, different ranks.
    assert general_wheel_117 is source_117.select_higher_rank(general_wheel_117)
    assert general_wheel_117 is source_117.select_higher_rank(
        general_wheel_117, prefer_older_binary=True
    )
    assert specific_wheel_117 is general_wheel_117.select_higher_rank(specific_wheel_117)
    assert specific_wheel_117 is general_wheel_117.select_higher_rank(
        specific_wheel_117, prefer_older_binary=True
    )

    # Same rank, same version.
    assert source_117 is source_117.select_higher_rank(source_117)
    assert source_117 is source_117.select_higher_rank(source_117, prefer_older_binary=True)

    # Rank and version covariant
    assert specific_wheel_118 is source_117.select_higher_rank(specific_wheel_118)
    assert specific_wheel_118 is source_117.select_higher_rank(
        specific_wheel_118, prefer_older_binary=True
    )

    # Rank and version contravariant all wheel.
    assert general_wheel_118 is general_wheel_118.select_higher_rank(specific_wheel_117)
    assert general_wheel_118 is general_wheel_118.select_higher_rank(
        specific_wheel_117, prefer_older_binary=True
    )

    # Rank and version contravariant mixed source and wheel.
    assert source_118 is source_118.select_higher_rank(specific_wheel_117)
    assert specific_wheel_117 is source_118.select_higher_rank(
        specific_wheel_117, prefer_older_binary=True
    )


@pytest.fixture
def cyclic_resolve():
    # type: () -> LockedResolve
    return LockedResolve(
        platform_tag=Tag("cp37", "cp37m", "linux_x86_64"),
        locked_requirements=locked_requirements(
            LockedRequirement.create(
                pin=pin("A", "1.0.0"),
                artifact=artifact("file:///repo/A-1.0.0.tar.gz", "sha256", "cafebabe"),
                requires_dists=requirements("B>=2", "C"),
            ),
            LockedRequirement.create(
                pin=pin("B", "2.0.0"),
                artifact=artifact("file:///repo/B-2.0-py2.py3-none-any.whl", "sha256", "cafebabe"),
                requires_dists=requirements("C"),
            ),
            LockedRequirement.create(
                pin=pin("C", "3.0.0"),
                artifact=artifact("file:///repo/C-3.0.0.tar.gz", "sha256", "cafebabe"),
                requires_dists=requirements("A~=1.0"),
            ),
        ),
    )


def test_transitive(
    current_target,  # type: Target
    cyclic_resolve,  # type: LockedResolve
):
    # type: (...) -> None

    assert_resolved(
        cyclic_resolve.resolve(current_target, [req("A")]),
        DownloadableArtifact.create(
            pin=pin("A", "1.0.0"),
            artifact=artifact("file:///repo/A-1.0.0.tar.gz", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements("A"),
        ),
        DownloadableArtifact.create(
            pin=pin("B", "2.0.0"),
            artifact=artifact("file:///repo/B-2.0-py2.py3-none-any.whl", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements(),
        ),
        DownloadableArtifact.create(
            pin=pin("C", "3.0.0"),
            artifact=artifact("file:///repo/C-3.0.0.tar.gz", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements(),
        ),
    )

    assert_resolved(
        cyclic_resolve.resolve(current_target, [req("A")], transitive=False),
        DownloadableArtifact.create(
            pin=pin("A", "1.0.0"),
            artifact=artifact("file:///repo/A-1.0.0.tar.gz", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements("A"),
        ),
    )

    assert_resolved(
        cyclic_resolve.resolve(current_target, [req("B")], transitive=False),
        DownloadableArtifact.create(
            pin=pin("B", "2.0.0"),
            artifact=artifact("file:///repo/B-2.0-py2.py3-none-any.whl", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements("B"),
        ),
    )

    assert_resolved(
        cyclic_resolve.resolve(current_target, [req("C")], transitive=False),
        DownloadableArtifact.create(
            pin=pin("C", "3.0.0"),
            artifact=artifact("file:///repo/C-3.0.0.tar.gz", "sha256", "cafebabe"),
            satisfied_direct_requirements=requirements("C"),
        ),
    )


def test_multiple_errors(
    current_target,  # type: Target
    cyclic_resolve,  # type: LockedResolve
):
    # type: (...) -> None

    assert_error(
        cyclic_resolve.resolve(current_target, [req("A==1.0.1")], use_wheel=False),
        dedent(
            """\
            Failed to resolve all requirements for {target_description}:

            Configured with:
                build: True
                use_wheel: False

            Dependency on a not satisfied, 1 incompatible candidate found:
            1.) a 1 does not satisfy the following requirements:
                ==1.0.1 (via: A==1.0.1)
            
            Dependency on b not satisfied, 1 incompatible candidate found:
            1.) b 2 (via: A==1.0.1 -> B>=2) does not have any compatible artifacts:
                file:///repo/B-2.0-py2.py3-none-any.whl
            """
        ).format(target_description=current_target.render_description()),
    )


def test_resolved():
    # type: () -> None

    def assert_resolved(
        expected_target_specificity,  # type: float
        supported_tag_count,  # type: int
        artifact_ranks,  # type: Iterable[int]
    ):
        # type: (...) -> None
        direct_requirements = requirements("foo")

        resolved_artifacts = tuple(
            resolved_artifact(
                project_name="foo",
                version=str(rank),
                artifact_basename="foo-{}.tar.gz".format(rank),
                rank_value=rank,
            )
            for rank in artifact_ranks
        )

        downloadable_artifacts = tuple(
            DownloadableArtifact.create(
                pin=resolved_art.locked_requirement.pin,
                artifact=resolved_art.artifact,
                satisfied_direct_requirements=direct_requirements,
            )
            for resolved_art in resolved_artifacts
        )

        target = CompletePlatform.create(
            MarkerEnvironment(),
            CompatibilityTags.from_strings(
                "py3-none-manylinux_2_{glibc_minor}_x86_64".format(glibc_minor=glibc_minor)
                for glibc_minor in range(supported_tag_count)
            ),
        )

        class DevNullFingerprinter(object):
            def fingerprint(self, _artifacts):
                # type: (Iterable[PartialArtifact]) -> Iterator[FileArtifact]
                return iter(())

        locked_resolve = LockedResolve.create(
            resolved_requirements=(),
            dist_metadatas=(),
            fingerprinter=DevNullFingerprinter(),
        )
        assert Resolved(
            target_specificity=expected_target_specificity,
            downloadable_artifacts=downloadable_artifacts,
            source=locked_resolve,
        ) == Resolved.create(
            target=target,
            direct_requirements=direct_requirements,
            resolved_artifacts=resolved_artifacts,
            source=locked_resolve,
        )

    # For tag ranks of 1, 2, 1 should rank 100% target specific (best match) and 2 should rank 0%
    # (worst match / universal)
    assert_resolved(expected_target_specificity=1.0, supported_tag_count=2, artifact_ranks=[1])
    assert_resolved(expected_target_specificity=0.0, supported_tag_count=2, artifact_ranks=[2])

    # For tag ranks of 1, 2, 3, 2 lands in the middle and should be 50% target specific.
    assert_resolved(expected_target_specificity=0.5, supported_tag_count=3, artifact_ranks=[2])


def test_file_artifact():
    # type: () -> None

    artifact = Artifact.from_url(
        url="file:///repo/A-1.0.0.tar.gz",
        fingerprint=Fingerprint(algorithm="md5", hash="foo"),
    )
    assert isinstance(artifact, FileArtifact)
    assert "A-1.0.0.tar.gz" == artifact.filename
    assert artifact.is_source
    assert frozenset() == frozenset(artifact.parse_tags())

    artifact = Artifact.from_url(
        url="https://example.org/ansicolors-1.1.7-py2.py3-none-any.whl",
        fingerprint=Fingerprint(algorithm="sha1", hash="foo"),
    )
    assert isinstance(artifact, FileArtifact)
    assert "ansicolors-1.1.7-py2.py3-none-any.whl" == artifact.filename
    assert not artifact.is_source
    assert frozenset((Tag("py2", "none", "any"), Tag("py3", "none", "any"))) == frozenset(
        artifact.parse_tags()
    )


def test_vcs_artifact():
    # type: () -> None

    artifact = Artifact.from_url(
        url="git+https://github.com/pantsbuild/pex",
        fingerprint=Fingerprint(algorithm="md5", hash="bar"),
    )
    assert isinstance(artifact, VCSArtifact)
    assert VCS.Git is artifact.vcs
    assert artifact.is_source
    assert "pex @ git+https://github.com/pantsbuild/pex" == artifact.as_unparsed_requirement(
        ProjectName("pex")
    )

    artifact = Artifact.from_url(
        url="hg+https://github.com/pantsbuild/pex#egg=pex&subdirectory=.",
        fingerprint=Fingerprint(algorithm="sha1", hash="bar"),
    )
    assert isinstance(artifact, VCSArtifact)
    assert VCS.Mercurial is artifact.vcs
    assert artifact.is_source
    assert (
        "hg+https://github.com/pantsbuild/pex#egg=pex&subdirectory=."
        == artifact.as_unparsed_requirement(ProjectName("pex"))
    )


def test_locked_requirement_mixed_artifacts_issue_2150():
    # type: () -> None

    file_artifact = FileArtifact(
        url="https://host/project.whl",
        fingerprint=Fingerprint(algorithm="md5", hash="foo"),
        verified=False,
        filename="a.whl",
    )
    vcs_artifact = VCSArtifact(
        url="git+https://host/a/project",
        fingerprint=Fingerprint(algorithm="sha1", hash="bar"),
        verified=False,
        vcs=VCS.Git,
    )
    local_project_artifact = LocalProjectArtifact(
        url="file:///tmp/project",
        fingerprint=Fingerprint(algorithm="sha256", hash="baz"),
        verified=False,
        directory="/tmp/project",
    )
    locked_requirement = LockedRequirement.create(
        pin=Pin(ProjectName("project"), Version("0.1.0")),
        artifact=local_project_artifact,
        additional_artifacts=(file_artifact, vcs_artifact),
    )
    assert [local_project_artifact, vcs_artifact, file_artifact] == list(
        locked_requirement.iter_artifacts()
    ), (
        "Expected the primary artifact to be 1st, and then the additional artifacts sorted by URL "
        "(git+https:// comes before https://)."
    )


def test_locked_resolve_same_pins_mixed_primary_artifacts_issue_2150():
    # type: () -> None

    pin = Pin(ProjectName("project"), Version("0.1.0"))
    file_artifact_requirement = LockedRequirement.create(
        pin,
        FileArtifact(
            url="https://host/project.whl",
            fingerprint=Fingerprint(algorithm="md5", hash="foo"),
            verified=False,
            filename="a.whl",
        ),
    )
    vcs_artifact_requirement = LockedRequirement.create(
        pin,
        VCSArtifact(
            url="git+https://host/a/project",
            fingerprint=Fingerprint(algorithm="sha1", hash="bar"),
            verified=False,
            vcs=VCS.Git,
        ),
    )
    local_project_artifact_requirement = LockedRequirement.create(
        pin,
        LocalProjectArtifact(
            url="file:///tmp/project",
            fingerprint=Fingerprint(algorithm="sha256", hash="baz"),
            verified=False,
            directory="/tmp/project",
        ),
    )
    locked_requirements = [
        file_artifact_requirement,
        vcs_artifact_requirement,
        local_project_artifact_requirement,
    ]
    locked_resolve = LockedResolve(locked_requirements=SortedTuple(locked_requirements))
    assert [
        local_project_artifact_requirement,
        vcs_artifact_requirement,
        file_artifact_requirement,
    ] == list(locked_resolve.locked_requirements)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.cli.command import BuildTimeCommand
from pex.cli.commands.interpreter import Interpreter
from pex.cli.commands.lock import Lock
from pex.cli.commands.venv import Venv
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Type


def all_commands():
    # type: () -> Iterable[Type[BuildTimeCommand]]
    return Interpreter, Lock, Venv



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from argparse import Action, ArgumentTypeError, Namespace, _ActionsContainer

from pex import pex_warnings
from pex.argparse import HandleBoolAction
from pex.network_configuration import NetworkConfiguration
from pex.orderedset import OrderedSet
from pex.pip.version import PipVersion
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.resolve.resolver_configuration import (
    PYPI,
    LockRepositoryConfiguration,
    PexRepositoryConfiguration,
    PipConfiguration,
    ReposConfiguration,
    ResolverVersion,
)
from pex.result import Error
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Optional, Union


class _ManylinuxAction(Action):
    def __init__(self, *args, **kwargs):
        kwargs["nargs"] = "?"
        super(_ManylinuxAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_str=None):
        if option_str.startswith("--no"):
            setattr(namespace, self.dest, None)
        elif value.startswith("manylinux"):
            setattr(namespace, self.dest, value)
        else:
            raise ArgumentTypeError(
                "Please specify a manylinux standard; ie: --manylinux=manylinux1. "
                "Given {}".format(value)
            )


class _HandleTransitiveAction(Action):
    def __init__(self, *args, **kwargs):
        kwargs["nargs"] = 0
        super(_HandleTransitiveAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_str=None):
        setattr(namespace, self.dest, option_str == "--transitive")


def register(
    parser,  # type: _ActionsContainer
    include_pex_repository=False,  # type: bool
    include_lock=False,  # type: bool
):
    # type: (...) -> None
    """Register resolver configuration options with the given parser.

    :param parser: The parser to register resolver configuration options with.
    :param include_pex_repository: Whether to include the `--pex-repository` option.
    :param include_lock: Whether to include the `--lock` option.
    """

    default_resolver_configuration = PipConfiguration()
    parser.add_argument(
        "--resolver-version",
        dest="resolver_version",
        default=default_resolver_configuration.resolver_version,
        choices=ResolverVersion.values(),
        type=ResolverVersion.for_value,
        help=(
            "The dependency resolver version to use. Read more at "
            "https://pip.pypa.io/en/stable/user_guide/#resolver-changes-2020"
        ),
    )
    parser.add_argument(
        "--pip-version",
        dest="pip_version",
        default=str(default_resolver_configuration.version),
        choices=["latest", "vendored"] + [str(value) for value in PipVersion.values()],
        help=(
            "The version of Pip to use for resolving dependencies. The `latest` version refers to "
            "the latest version in this list ({latest}) which is not necessarily the latest Pip "
            "version released on PyPI.".format(latest=PipVersion.LATEST)
        ),
    )
    parser.add_argument(
        "--allow-pip-version-fallback",
        "--no-allow-pip-version-fallback",
        dest="allow_pip_version_fallback",
        default=default_resolver_configuration.allow_version_fallback,
        action=HandleBoolAction,
        help=(
            "Whether to allow --pip-version to be ignored if the requested version is not "
            "compatible with all of the selected interpreters. If fallback is allowed, a warning "
            "will be emitted when fallback is necessary. If fallback is not allowed, Pex will fail "
            "fast indicating the problematic selected interpreters."
        ),
    )

    register_repos_options(parser)
    register_network_options(parser)

    parser.add_argument(
        "--cache-ttl",
        metavar="DEPRECATED",
        default=None,
        type=int,
        help="Deprecated: No longer used.",
    )
    parser.add_argument(
        "-H",
        "--header",
        dest="headers",
        metavar="DEPRECATED",
        default=None,
        type=str,
        action="append",
        help="Deprecated: No longer used.",
    )

    repository_choice = (
        parser.add_mutually_exclusive_group() if include_pex_repository and include_lock else parser
    )
    if include_pex_repository:
        repository_choice.add_argument(
            "--pex-repository",
            dest="pex_repository",
            metavar="FILE",
            default=None,
            type=str,
            help=(
                "Resolve requirements from the given PEX file instead of from --index servers, "
                "--find-links repos or a --lock file."
            ),
        )
    if include_lock:
        repository_choice.add_argument(
            "--lock",
            dest="lock",
            metavar="FILE",
            default=None,
            type=str,
            help=(
                "Resolve requirements from the given lock file created by Pex instead of from "
                "--index servers, --find-links repos or a --pex-repository. If no requirements are "
                "specified, will install the entire lock."
            ),
        )
        register_lock_options(parser)

    parser.add_argument(
        "--pre",
        "--no-pre",
        dest="allow_prereleases",
        default=default_resolver_configuration.allow_prereleases,
        action=HandleBoolAction,
        help="Whether to include pre-release and development versions of requirements.",
    )
    parser.add_argument(
        "--wheel",
        "--binary",
        "--no-wheel",
        "--no-use-wheel",
        "--no-binary",
        "--no-use-binary",
        dest="allow_wheels",
        default=default_resolver_configuration.allow_wheels,
        action=HandleBoolAction,
        help="Whether to allow binary distributions.",
    )
    parser.add_argument(
        "--build",
        "--no-build",
        dest="allow_builds",
        default=default_resolver_configuration.allow_builds,
        action=HandleBoolAction,
        help="Whether to allow building of distributions from source.",
    )
    parser.add_argument(
        "--prefer-wheel",
        "--prefer-binary",
        "--no-prefer-wheel",
        "--no-prefer-binary",
        dest="prefer_older_binary",
        default=default_resolver_configuration.prefer_older_binary,
        action=HandleBoolAction,
        help=(
            "Whether to prefer older binary distributions to newer source distributions (prefer "
            "not building wheels)."
        ),
    )
    parser.add_argument(
        "--force-pep517",
        "--use-pep517",
        "--no-use-pep517",
        dest="use_pep517",
        default=default_resolver_configuration.use_pep517,
        action=HandleBoolAction,
        help=(
            "Whether to force use of PEP 517 for building source distributions into wheels ("
            "https://www.python.org/dev/peps/pep-0518) or force direct invocation of"
            "`setup.py bdist_wheel` (which requires all source distributions have a `setup.py` "
            "based build). Defaults to using PEP-517 only when a `pyproject.toml` file is present "
            "with a `build-system` section. If PEP-517 is forced (--use-pep517 is passed) and no "
            "`pyproject.toml` file is present or one is but does not have a `build-system` section "
            "defined, then the build is executed as if a `pyproject.toml` was present with a "
            '`build-system` section comprised of `requires = ["setuptools>=40.8.0", "wheel"]` and '
            '`build-backend = "setuptools.build_meta:__legacy__"`.'
        ),
    )
    parser.add_argument(
        "--build-isolation",
        "--no-build-isolation",
        dest="build_isolation",
        default=default_resolver_configuration.build_isolation,
        action=HandleBoolAction,
        help=(
            "Disable `sys.path` isolation when building a modern source distribution. Build "
            "dependencies specified by PEP 518 (https://www.python.org/dev/peps/pep-0518) must "
            "already be installed on the `sys.path` if this option is used."
        ),
    )
    parser.add_argument(
        "--transitive",
        "--no-transitive",
        "--intransitive",
        dest="transitive",
        default=default_resolver_configuration.transitive,
        action=_HandleTransitiveAction,
        help="Whether to transitively resolve requirements.",
    )
    register_max_jobs_option(parser)
    parser.add_argument(
        "--preserve-pip-download-log",
        "--no-preserve-pip-download-log",
        default=default_resolver_configuration.preserve_log,
        action=HandleBoolAction,
        help="Preserve the `pip download` log and print its location to stderr.",
    )


def register_lock_options(parser):
    # type: (_ActionsContainer) -> None
    """Register lock options with the given parser.

    :param parser: The parser to register lock configuration options with.
    """
    parser.add_argument(
        "--path-mapping",
        dest="path_mappings",
        action="append",
        default=[],
        type=str,
        help=(
            "A mapping of the form `NAME|PATH|DESCRIPTION` of a logical name to a concrete local "
            "absolute path with an optional description. Can be specified multiple times. The "
            "mapping must include the pipe (`|`) separated name and absolute path components, but "
            "the trailing pipe-separated description is optional. The mapping is used when "
            "creating, and later reading, lock files to ensure the lock file created on one "
            "machine can be used on another with a potentially different realization of various "
            "paths used in the resolve. A typical example is a find-links repo. This might be "
            "provided on the file-system via a network mount instead of via an HTTP(S) server and "
            "that network mount may be at different absolute paths on different machines. "
            "Classically, it may be in a user's home directory; whose path will vary from user to "
            "user."
        ),
    )


def register_repos_options(parser):
    # type: (_ActionsContainer) -> None
    """Register repos configuration options with the given parser.

    :param parser: The parser to register repos configuration options with.
    """
    parser.add_argument(
        "--pypi",
        "--no-pypi",
        "--no-index",
        dest="pypi",
        action=HandleBoolAction,
        default=True,
        help="Whether to use PyPI to resolve dependencies.",
    )
    parser.add_argument(
        "-f",
        "--find-links",
        "--repo",
        metavar="PATH/URL",
        action="append",
        dest="find_links",
        type=str,
        help="Additional repository path (directory or URL) to look for requirements.",
    )
    parser.add_argument(
        "-i",
        "--index",
        "--index-url",
        metavar="URL",
        action="append",
        dest="indexes",
        type=str,
        help="Additional cheeseshop indices to use to satisfy requirements.",
    )


def register_network_options(parser):
    # type: (_ActionsContainer) -> None
    """Register network configuration options with the given parser.

    :param parser: The parser to register network configuration options with.
    """
    default_resolver_configuration = PipConfiguration()
    default_network_configuration = default_resolver_configuration.network_configuration
    parser.add_argument(
        "--retries",
        default=default_network_configuration.retries,
        type=int,
        help="Maximum number of retries each connection should attempt.",
    )
    parser.add_argument(
        "--timeout",
        metavar="SECS",
        default=default_network_configuration.timeout,
        type=int,
        help="Set the socket timeout in seconds.",
    )
    parser.add_argument(
        "--proxy",
        type=str,
        default=default_network_configuration.proxy,
        help="Specify a proxy in the form http(s)://[user:passwd@]proxy.server:port.",
    )
    parser.add_argument(
        "--cert",
        metavar="PATH",
        type=str,
        default=default_network_configuration.cert,
        help="Path to alternate CA bundle.",
    )
    parser.add_argument(
        "--client-cert",
        metavar="PATH",
        type=str,
        default=default_network_configuration.client_cert,
        help=(
            "Path to an SSL client certificate which should be a single file containing the "
            "private key and the certificate in PEM format."
        ),
    )


def register_max_jobs_option(parser):
    # type: (_ActionsContainer) -> None
    """Register the max jobs configuration option with the given parser.

    :param parser: The parser to register the max job option with.
    """
    default_resolver_configuration = PipConfiguration()
    parser.add_argument(
        "-j",
        "--jobs",
        metavar="JOBS",
        dest="max_jobs",
        type=int,
        default=default_resolver_configuration.max_jobs,
        help=(
            "The maximum number of parallel jobs to use when resolving, building and "
            "installing distributions. You might want to increase the maximum number of "
            "parallel jobs to potentially improve the latency of the pex creation process at "
            "the expense of other processes on your system."
        ),
    )


class InvalidConfigurationError(Exception):
    """Indicates an invalid resolver configuration."""


if TYPE_CHECKING:
    ResolverConfiguration = Union[
        LockRepositoryConfiguration, PexRepositoryConfiguration, PipConfiguration
    ]


def configure(options):
    # type: (Namespace) -> ResolverConfiguration
    """Creates a resolver configuration from options registered by `register`.

    :param options: The resolver configuration options.
    :raise: :class:`InvalidConfigurationError` if the resolver configuration is invalid.
    """

    pex_repository = getattr(options, "pex_repository", None)
    lock = getattr(options, "lock", None)
    if pex_repository and (options.indexes or options.find_links):
        raise InvalidConfigurationError(
            'The "--pex-repository" option cannot be used together with the "--index" or '
            '"--find-links" options.'
        )

    if pex_repository:
        return PexRepositoryConfiguration(
            pex_repository=pex_repository,
            network_configuration=create_network_configuration(options),
            transitive=options.transitive,
        )
    pip_configuration = create_pip_configuration(options)
    if lock:
        return LockRepositoryConfiguration(
            parse_lock=lambda: parse_lockfile(options, lock_file_path=lock),
            lock_file_path=lock,
            pip_configuration=pip_configuration,
        )
    return pip_configuration


def create_pip_configuration(options):
    # type: (Namespace) -> PipConfiguration
    """Creates a Pip configuration from options registered by `register`.

    :param options: The Pip resolver configuration options.
    """

    if options.cache_ttl:
        pex_warnings.warn("The --cache-ttl option is deprecated and no longer has any effect.")
    if options.headers:
        pex_warnings.warn("The --header option is deprecated and no longer has any effect.")

    repos_configuration = create_repos_configuration(options)

    if options.pip_version == "latest":
        pip_version = PipVersion.LATEST
    elif options.pip_version == "vendored":
        pip_version = PipVersion.VENDORED
    else:
        pip_version = PipVersion.for_value(options.pip_version)

    return PipConfiguration(
        resolver_version=options.resolver_version,
        repos_configuration=repos_configuration,
        network_configuration=create_network_configuration(options),
        allow_prereleases=options.allow_prereleases,
        allow_wheels=options.allow_wheels,
        allow_builds=options.allow_builds,
        prefer_older_binary=options.prefer_older_binary,
        use_pep517=options.use_pep517,
        build_isolation=options.build_isolation,
        transitive=options.transitive,
        max_jobs=get_max_jobs_value(options),
        preserve_log=options.preserve_pip_download_log,
        version=pip_version,
        allow_version_fallback=options.allow_pip_version_fallback,
    )


def create_repos_configuration(options):
    # type: (Namespace) -> ReposConfiguration
    """Creates a repos configuration from options registered by `register_repos_options`.

    :param options: The Pip resolver configuration options.
    """
    indexes = OrderedSet(
        ([PYPI] if options.pypi else []) + (options.indexes or [])
    )  # type: OrderedSet[str]
    find_links = OrderedSet(options.find_links or ())  # type: OrderedSet[str]
    return ReposConfiguration.create(indexes=tuple(indexes), find_links=tuple(find_links))


def create_network_configuration(options):
    # type: (Namespace) -> NetworkConfiguration
    """Creates a network configuration from options registered by `register_network_options`.

    :param options: The Pip resolver configuration options.
    """
    return NetworkConfiguration(
        retries=options.retries,
        timeout=options.timeout,
        proxy=options.proxy,
        cert=options.cert,
        client_cert=options.client_cert,
    )


def get_max_jobs_value(options):
    # type: (Namespace) -> int
    """Retrieves the max jobs value from the option registered by `register_max_jobs_option`.

    :param options: The max jobs configuration option.
    """
    return cast(int, options.max_jobs)


def _parse_path_mapping(path_mapping):
    # type: (str) -> PathMapping
    components = path_mapping.split("|", 2)
    if len(components) < 2:
        raise ArgumentTypeError(
            "A path mapping must be of the form `NAME|PATH` with an optional trailing "
            "`|DESCRIPTION`, given: {path_mapping}.\n"
            "For example: `FL|/path/to/local/find-links/repo/directory` indicates that find-links "
            "requirements or URLs starting with `/path/to/local/find-links/repo/directory` should "
            "have that absolute root path replaced with the `${{FL}}` placeholder name.\n"
            "Alternatively, you could use the form with a trailing description to make it more "
            "clear what value should be substituted for `${{FL}}` when the mapping is later read, "
            "e.g.: `FL|/local/path|The local find-links repo path`."
            "".format(path_mapping=path_mapping)
        )
    name, path = components[:2]
    description = components[2] if len(components) == 3 else None
    return PathMapping(path=path, name=name, description=description)


def get_path_mappings(options):
    # type: (Namespace) -> PathMappings
    """Retrieves the PathMappings value from the options registered by `register_lock_options`.

    :param options: The lock configuration options.
    """
    return PathMappings(
        mappings=tuple(_parse_path_mapping(path_mapping) for path_mapping in options.path_mappings)
    )


def parse_lockfile(
    options,  # type: Namespace
    lock_file_path=None,  # type: Optional[str]
):
    # type: (...) -> Union[Lockfile, Error]
    path = lock_file_path or options.lock
    path_mappings = get_path_mappings(options)
    with TRACER.timed("Parsing lock {lockfile}".format(lockfile=path)):
        try:
            return json_codec.load(path, path_mappings=path_mappings)
        except json_codec.PathMappingError as e:
            return Error(
                "The lockfile at {path} requires specifying {prefix}"
                "'--path-mapping' {values} for: {required_paths}\n"
                "Given {given_mappings_verbiage}\n"
                "{maybe_path_mappings}"
                "Which left the following path mappings unspecified:\n"
                "{unspecified_paths}\n"
                "\n"
                "To fix, add command line options for:\n{examples}".format(
                    path=path,
                    prefix="" if len(e.required_path_mappings) > 1 else "a ",
                    values="values" if len(e.required_path_mappings) > 1 else "value",
                    required_paths=", ".join(sorted(e.required_path_mappings)),
                    given_mappings_verbiage="the following path mappings:"
                    if path_mappings.mappings
                    else "no path mappings.",
                    maybe_path_mappings="{path_mappings}\n".format(
                        path_mappings="\n".join(
                            sorted(
                                "--path-mapping '{mapping}'".format(
                                    mapping="|".join((mapping.name, mapping.path))
                                )
                                for mapping in path_mappings.mappings
                            )
                        )
                    )
                    if path_mappings.mappings
                    else "",
                    unspecified_paths="\n".join(
                        sorted(
                            (
                                "{path}: {description}".format(path=path, description=description)
                                if description
                                else path
                            )
                            for path, description in e.required_path_mappings.items()
                            if path in e.unspecified_paths
                        )
                    ),
                    examples="\n".join(
                        sorted(
                            "--path-mapping '{path}|<path of {path}>'".format(path=path)
                            for path in e.required_path_mappings
                            if path in e.unspecified_paths
                        )
                    ),
                )
            )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import itertools
import os
import shutil
import subprocess

import pytest

from pex.cli.testing import run_pex3
from pex.common import safe_open
from pex.fetcher import URLFetcher
from pex.resolve.locked_resolve import Artifact, FileArtifact, LockedRequirement
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolver_configuration import PipConfiguration
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterator


def iter_requirements(lf):
    # type: (Lockfile) -> Iterator[LockedRequirement]
    return itertools.chain.from_iterable(
        locked_resolve.locked_requirements for locked_resolve in lf.locked_resolves
    )


def assert_file_artifact(artifact):
    # type: (Artifact) -> FileArtifact
    assert isinstance(artifact, FileArtifact)
    return artifact


def iter_artifacts(lf):
    # type: (Lockfile) -> Iterator[FileArtifact]
    return itertools.chain.from_iterable(
        map(assert_file_artifact, locked_requirement.iter_artifacts())
        for locked_requirement in iter_requirements(lf)
    )


@pytest.fixture
def lock(tmpdir):
    # type: (Any) -> str
    return os.path.join(str(tmpdir), "lock")


@pytest.fixture
def lock_file(lock):
    # type: (str) -> Lockfile
    run_pex3(
        "lock", "create", "ansicolors==1.1.8", "--style", "universal", "-o", lock, "--indent", "2"
    ).assert_success()
    return json_codec.load(lock)


@pytest.fixture
def find_links(
    tmpdir,  # type: Any
    lock_file,  # type: Lockfile
):
    # type: (...) -> str

    url_fetcher = URLFetcher()
    find_links = os.path.join(str(tmpdir), "find_links")
    for artifact in iter_artifacts(lock_file):
        with url_fetcher.get_body_stream(artifact.url) as url_fp:
            with safe_open(os.path.join(find_links, artifact.filename), "wb") as fl_fp:
                shutil.copyfileobj(url_fp, fl_fp)

    # We need the current default --pip-version requirements for some tests that do PyPI offline
    # resolves.
    pip_version = PipConfiguration().version
    repository_pex = os.path.join(str(tmpdir), "repository.pex")
    run_pex_command(
        args=[
            pip_version.setuptools_requirement,
            pip_version.wheel_requirement,
            "--include-tools",
            "-o",
            repository_pex,
        ]
    ).assert_success()
    subprocess.check_call(
        args=[repository_pex, "repository", "extract", "-f", find_links], env=make_env(PEX_TOOLS=1)
    )
    return find_links


def test_lock_update_repo_migration_dry_run(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    result = run_pex3(
        "lock", "update", "--pin", "--no-pypi", "--find-links", find_links, lock, "--dry-run"
    )
    result.assert_success()
    assert (
        "Updates for lock generated by universal\n"
        "  Would update ansicolors 1.1.8 artifacts:\n"
        "    https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl -> "
        "file://{find_links}/ansicolors-1.1.8-py2.py3-none-any.whl\n"
        "    https://files.pythonhosted.org/packages/76/31/"
        "7faed52088732704523c259e24c26ce6f2f33fbeff2ff59274560c27628e/ansicolors-1.1.8.zip -> "
        "file://{find_links}/ansicolors-1.1.8.zip\n".format(find_links=find_links)
    ) == result.output


def test_lock_update_repo_migration_dry_run_path_mapping(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    result = run_pex3(
        "lock",
        "update",
        "--pin",
        "--no-pypi",
        "--find-links",
        find_links,
        lock,
        "--dry-run",
        "--path-mapping",
        "FOO|{find_links}".format(find_links=find_links),
    )
    result.assert_success()
    assert (
        "Updates for lock generated by universal\n"
        "  Would update ansicolors 1.1.8 artifacts:\n"
        "    https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl -> "
        "file://${FOO}/ansicolors-1.1.8-py2.py3-none-any.whl\n"
        "    https://files.pythonhosted.org/packages/76/31/"
        "7faed52088732704523c259e24c26ce6f2f33fbeff2ff59274560c27628e/ansicolors-1.1.8.zip -> "
        "file://${FOO}/ansicolors-1.1.8.zip\n"
    ) == result.output


def test_lock_update_repo_migration_nominal(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    artifacts_by_filename = {artifact.filename: artifact for artifact in iter_artifacts(lock_file)}

    run_pex3(
        "lock", "update", "--pin", "--no-pypi", "--find-links", find_links, lock
    ).assert_success()

    for updated_artifact in iter_artifacts(json_codec.load(lock)):
        original_artifact = artifacts_by_filename.pop(updated_artifact.filename)
        assert original_artifact.fingerprint == updated_artifact.fingerprint
        assert original_artifact.url != updated_artifact.url
        assert (
            "file://{}".format(os.path.join(find_links, original_artifact.filename))
            == updated_artifact.url
        )
    assert not artifacts_by_filename


def test_lock_update_repo_migration_corrupted(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    with open(os.path.join(find_links, "ansicolors-1.1.8.zip"), "ab") as fp:
        fp.write(b"changed")

    result = run_pex3(
        "lock",
        "update",
        "--pin",
        "--no-pypi",
        "--find-links",
        find_links,
        lock,
        "--dry-run",
        "--path-mapping",
        "FOO|{find_links}".format(find_links=find_links),
    )
    result.assert_failure()
    assert (
        "Updates for lock generated by universal\n"
        "  Would update ansicolors 1.1.8 artifacts:\n"
        "    https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl -> "
        "file://${FOO}/ansicolors-1.1.8-py2.py3-none-any.whl\n"
        "    https://files.pythonhosted.org/packages/76/31/"
        "7faed52088732704523c259e24c26ce6f2f33fbeff2ff59274560c27628e/"
        "ansicolors-1.1.8.zip"
        "#sha256:99f94f5e3348a0bcd43c82e5fc4414013ccc19d70bd939ad71e0133ce9c372e0 -> "
        "file://${FOO}/ansicolors-1.1.8.zip"
        "#sha256:9c872ada674e45fe740ebd4d00267617875e76d9bec4b3ce2e194e83661680dc\n"
    ) == result.output
    assert (
        "Detected fingerprint changes in the following locked project for lock generated by "
        "universal!\n"
        "ansicolors 1.1.8\n"
    ) == result.error


def test_lock_update_repo_migration_artifacts_removed(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    os.unlink(os.path.join(find_links, "ansicolors-1.1.8-py2.py3-none-any.whl"))
    result = run_pex3(
        "lock",
        "update",
        "--pin",
        "--no-pypi",
        "--find-links",
        find_links,
        lock,
        "--dry-run",
        "--path-mapping",
        "FOO|{find_links}".format(find_links=find_links),
    )
    result.assert_success()
    assert (
        "Updates for lock generated by universal\n"
        "  Would update ansicolors 1.1.8 artifacts:\n"
        "    https://files.pythonhosted.org/packages/76/31/"
        "7faed52088732704523c259e24c26ce6f2f33fbeff2ff59274560c27628e/ansicolors-1.1.8.zip -> "
        "file://${FOO}/ansicolors-1.1.8.zip\n"
        "    - https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl\n"
    ) == result.output


def test_lock_update_repo_migration_artifacts_added(
    lock,  # type: str
    lock_file,  # type: Lockfile
    find_links,  # type: str
):
    # type: (...) -> None

    shutil.copy(
        os.path.join(find_links, "ansicolors-1.1.8.zip"),
        os.path.join(find_links, "ansicolors-1.1.8.tar.gz"),
    )
    result = run_pex3(
        "lock",
        "update",
        "--pin",
        "--no-pypi",
        "--find-links",
        find_links,
        lock,
        "--dry-run",
        "--path-mapping",
        "FOO|{find_links}".format(find_links=find_links),
    )
    result.assert_success()
    assert (
        "Updates for lock generated by universal\n"
        "  Would update ansicolors 1.1.8 artifacts:\n"
        "    + file://${FOO}/ansicolors-1.1.8.tar.gz\n"
        "    https://files.pythonhosted.org/packages/53/18/"
        "a56e2fe47b259bb52201093a3a9d4a32014f9d85071ad07e9d60600890ca/"
        "ansicolors-1.1.8-py2.py3-none-any.whl -> "
        "file://${FOO}/ansicolors-1.1.8-py2.py3-none-any.whl\n"
        "    https://files.pythonhosted.org/packages/76/31/"
        "7faed52088732704523c259e24c26ce6f2f33fbeff2ff59274560c27628e/ansicolors-1.1.8.zip -> "
        "file://${FOO}/ansicolors-1.1.8.zip\n"
    ) == result.output, result.output



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess

from pex.testing import PY38, ensure_python_interpreter, make_env, pex_project_dir, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def assert_venv_runtime_env_vars_ignored_during_create(
    tmpdir,  # type: Any
    pex_venv,  # type: bool
):
    # type: (...) -> None

    pex_pex = os.path.join(str(tmpdir), "pex.pex")
    args = [pex_project_dir(), "-c", "pex", "-o", pex_pex, "--no-strip-pex-env", "--disable-cache"]
    if pex_venv:
        args.append("--venv")
    run_pex_command(args=args).assert_success()

    py38 = ensure_python_interpreter(PY38)
    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    subprocess.check_call(
        args=[
            py38,
            pex_pex,
            "lock",
            "create",
            "ansicolors==1.1.8",
            "-o",
            lock,
            "--pex-root",
            pex_root,
        ],
        env=make_env(PEX_SCRIPT="pex3"),
    )
    ansicolors_path = (
        subprocess.check_output(
            args=[
                py38,
                pex_pex,
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--lock",
                lock,
                "--",
                "-c",
                "import colors; print(colors.__file__)",
            ]
        )
        .decode("utf-8")
        .strip()
    )
    assert ansicolors_path.startswith(pex_root)


def test_venv_runtime_env_vars_ignored_during_create_nested(tmpdir):
    # type: (Any) -> None

    # N.B.: The venv being created here is the internal Pip venv Pex uses to execute Pip. It's that
    # venv that used to get PEX env vars like PEX_MODULE and PEX_SCRIPT sealed in at creation time.
    assert_venv_runtime_env_vars_ignored_during_create(tmpdir, pex_venv=False)


def test_venv_runtime_env_vars_ignored_during_create_top_level(tmpdir):
    assert_venv_runtime_env_vars_ignored_during_create(tmpdir, pex_venv=True)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.third_party.packaging import markers
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Dict


def evaluate_marker(
    expression,  # type: str
    environment,  # type: Dict[str, str]
):
    # type: (...) -> bool
    markers.default_environment = environment.copy
    return cast(bool, markers.Marker(expression).evaluate())


def test_platform_marker_environment():
    # type: () -> None
    platform = Platform.create("linux-x86_64-cp-37-cp37m")
    marker_environment = MarkerEnvironment.from_platform(platform)
    env = marker_environment.as_dict()

    def assert_known_marker(expression):
        # type: (str) -> None
        assert evaluate_marker(expression, env)

    assert_known_marker("python_version == '3.7'")
    assert_known_marker("implementation_name == 'cpython'")
    assert_known_marker("platform_system == 'Linux'")
    assert_known_marker("platform_machine == 'x86_64'")

    def assert_unknown_marker(expression):
        # type: (str) -> None
        with pytest.raises(markers.UndefinedEnvironmentName):
            evaluate_marker(expression, env)

    assert_unknown_marker("python_full_version == '3.7.10'")
    assert_unknown_marker("platform_release == '5.12.12-arch1-1'")
    assert_unknown_marker("platform_version == '#1 SMP PREEMPT Fri, 18 Jun 2021 21:59:22 +0000'")


def test_extended_platform_marker_environment():
    # type: () -> None
    platform = Platform.create("linux-x86_64-cp-3.10.1-cp310")
    marker_environment = MarkerEnvironment.from_platform(platform)
    env = marker_environment.as_dict()

    def assert_known_marker(expression):
        # type: (str) -> None
        assert evaluate_marker(expression, env)

    assert_known_marker("python_full_version == '3.10.1'")
    assert_known_marker("python_version == '3.10'")
    assert_known_marker("implementation_name == 'cpython'")
    assert_known_marker("platform_system == 'Linux'")
    assert_known_marker("platform_machine == 'x86_64'")

    def assert_unknown_marker(expression):
        # type: (str) -> None
        with pytest.raises(markers.UndefinedEnvironmentName):
            evaluate_marker(expression, env)

    assert_unknown_marker("platform_release == '5.12.12-arch1-1'")
    assert_unknown_marker("platform_version == '#1 SMP PREEMPT Fri, 18 Jun 2021 21:59:22 +0000'")


def test_platform_marker_environment_issue_1488():
    # type: () -> None

    def assert_platform_machine(
        expected,  # type: str
        platform,  # type: str
    ):
        marker_environment = MarkerEnvironment.from_platform(Platform.create(platform))
        assert expected == marker_environment.platform_machine

    assert_platform_machine("x86_64", "linux-x86_64-cp-37-cp37m")
    assert_platform_machine("x86_64", "manylinux1-x86_64-cp-37-cp37m")
    assert_platform_machine("x86_64", "manylinux2010-x86_64-cp-37-cp37m")
    assert_platform_machine("x86_64", "manylinux2014-x86_64-cp-37-cp37m")
    assert_platform_machine("x86_64", "manylinux_2_5-x86_64-cp-37-cp37m")
    assert_platform_machine("aarch64", "manylinux_2_77-aarch64-cp-37-cp37m")

    assert_platform_machine("x86_64", "macosx-10.15-x86_64-cp-38-m")
    assert_platform_machine("arm64", "macosx-11.0-arm64-cp-39-cp39")



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os
import re
import subprocess

from pex.testing import IS_PYPY, PY_VER, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_custom_prompt(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    venv_pex_name = "venv.pex"
    venv_pex = os.path.join(str(tmpdir), venv_pex_name)

    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "-o",
            venv_pex,
            "--seed",
            "verbose",
            "--venv",
        ]
    )
    result.assert_success()
    venv_dir = os.path.dirname(json.loads(result.output)["pex"])

    if PY_VER == (2, 7) or (IS_PYPY and PY_VER <= (3, 7)):
        # Neither CPython 2.7 not PyPy interpreters have (functioning) venv modules; so we create
        # their venvs with an old copy of virtualenv that does not surround the prompt with parens.
        expected_prompt = r"^{venv_pex_name}$".format(venv_pex_name=re.escape(venv_pex_name))
    elif PY_VER == (3, 5):
        # We can't set the prompt for CPython 3.5 so we expect a venv atomic_directory-style name.
        expected_prompt = r"^\({venv_dir}\.[0-9a-f]+\)$".format(
            venv_dir=re.escape(os.path.basename(venv_dir))
        )
    else:
        expected_prompt = r"^\({venv_pex_name}\)$".format(venv_pex_name=re.escape(venv_pex_name))

    output = subprocess.check_output(
        args=[
            "/usr/bin/env",
            "bash",
            "-c",
            "source {} && echo $PS1".format(os.path.join(venv_dir, "bin", "activate")),
        ],
        env=make_env(TERM="dumb", COLS=80),
    )
    assert re.match(expected_prompt, output.decode("utf-8").strip()) is not None



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.resolve.locker import CredentialedURL, Credentials, Netloc
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def test_credentials_redacted():
    # type: () -> None

    assert not Credentials("git").are_redacted()
    assert not Credentials("git", "").are_redacted()
    assert not Credentials("joe", "bob").are_redacted()

    assert Credentials("****").are_redacted()
    assert Credentials("joe", "****").are_redacted()


def test_basic_auth_rendering():
    # type: () -> None

    assert "git" == Credentials("git").render_basic_auth()
    assert "git:" == Credentials("git", "").render_basic_auth()
    assert "joe:bob" == Credentials("joe", "bob").render_basic_auth()


def test_host_port_rendering():
    # type: () -> None

    assert "example.com" == Netloc("example.com").render_host_port()
    assert "example.com:80" == Netloc("example.com", 80).render_host_port()


def test_strip_credentials():
    # type: () -> None

    def strip_credentials(
        url,  # type: str
        expected_credentials=None,  # type: Optional[Credentials]
    ):
        # type: (...) -> str
        credentialed_url = CredentialedURL.parse(url)
        assert expected_credentials == credentialed_url.credentials
        return str(credentialed_url.strip_credentials())

    assert "file:///a/file" == strip_credentials("file:///a/file")
    assert "http://example.com" == strip_credentials("http://example.com")
    assert "http://example.com:80" == strip_credentials("http://example.com:80")
    assert "http://example.com" == strip_credentials("http://joe@example.com", Credentials("joe"))
    assert "https://example.com:443" == strip_credentials(
        "https://joe@example.com:443", Credentials("joe")
    )
    assert "http://example.com" == strip_credentials(
        "http://joe:bob@example.com", Credentials("joe", "bob")
    )
    assert "phys://example.com:1137" == strip_credentials(
        "phys://joe:bob@example.com:1137", Credentials("joe", "bob")
    )
    assert "git://example.org" == strip_credentials("git://git@example.org", Credentials("git"))
    assert "git://example.org" == strip_credentials("git://****@example.org", Credentials("****"))


def test_inject_credentials():
    # type: () -> None

    def inject_credentials(
        url,  # type: str
        credentials,  # type: Optional[Credentials]
    ):
        # type: (...) -> str
        return str(CredentialedURL.parse(url).inject_credentials(credentials))

    assert "git://git@example.org" == inject_credentials("git://example.org", Credentials("git"))
    assert "git://git@example.org" == inject_credentials(
        "git://****@example.org", Credentials("git")
    )
    assert "http://joe:bob@example.org" == inject_credentials(
        "http://example.org", Credentials("joe", "bob")
    )
    assert "http://joe:bob@example.org" == inject_credentials(
        "http://joe:****@example.org", Credentials("joe", "bob")
    )



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.compatibility import PY3, commonpath, indent, to_bytes, to_unicode

unicode_string = (str,) if PY3 else (unicode,)  # type: ignore[name-defined]


def test_to_bytes():
    # type: () -> None
    assert isinstance(to_bytes(""), bytes)
    assert isinstance(to_bytes("abc"), bytes)
    assert isinstance(to_bytes(b"abc"), bytes)
    assert isinstance(to_bytes(u"abc"), bytes)
    assert isinstance(to_bytes(b"abc".decode("latin-1"), encoding=u"utf-8"), bytes)

    for bad_value in (123, None):
        with pytest.raises(ValueError):
            to_bytes(bad_value)  # type: ignore[type-var]


def test_to_unicode():
    # type: () -> None
    assert isinstance(to_unicode(""), unicode_string)
    assert isinstance(to_unicode("abc"), unicode_string)
    assert isinstance(to_unicode(b"abc"), unicode_string)
    assert isinstance(to_unicode(u"abc"), unicode_string)
    assert isinstance(to_unicode(u"abc".encode("latin-1"), encoding=u"latin-1"), unicode_string)

    for bad_value in (123, None):
        with pytest.raises(ValueError):
            to_unicode(bad_value)  # type: ignore[type-var]


def test_indent():
    # type: () -> None
    assert "  line1" == indent("line1", "  ")

    assert "  line1\n  line2" == indent("line1\nline2", "  ")
    assert "  line1\n  line2\n" == indent("line1\nline2\n", "  ")

    assert "  line1\n\n  line3" == indent("line1\n\nline3", "  ")
    assert "  line1\n \n  line3" == indent("line1\n \nline3", "  ")
    assert "  line1\n  \n  line3" == indent("line1\n\nline3", "  ", lambda line: True)


def test_commonpath_invalid():
    # type: () -> None

    with pytest.raises(ValueError):
        commonpath([])

    with pytest.raises(ValueError):
        commonpath(["a", "/a"])


def test_commonpath_single():
    # type: () -> None

    assert "" == commonpath([""])
    assert "/" == commonpath(["/"])
    assert "a" == commonpath(["a"])
    assert "/a" == commonpath(["/a"])


def test_commonpath_common():
    # type: () -> None

    assert "a" == commonpath(["a", "a"])
    assert "a" == commonpath(["a", "a/"])
    assert "a" == commonpath(["a", "a/b"])
    assert "a" == commonpath(["a/", "a/b"])
    assert "a" == commonpath(["a/c", "a/b"])

    assert "/a" == commonpath(["/a", "/a"])
    assert "/a" == commonpath(["/a", "/a/"])
    assert "/a" == commonpath(["/a", "/a/b"])
    assert "/a" == commonpath(["/a/", "/a/b"])
    assert "/a" == commonpath(["/a/c", "/a/b"])

    assert "a" == commonpath(["./a", "./a"])
    assert "a" == commonpath(["./a", "./a/"])
    assert "a" == commonpath(["./a", "./a/b"])
    assert "a" == commonpath(["./a/", "./a/b"])
    assert "a" == commonpath(["./a/c", "./a/b"])

    assert "../a" == commonpath(["../a", "../a"])
    assert "../a" == commonpath(["../a", "../a/"])
    assert "../a" == commonpath(["../a", "../a/b"])
    assert "../a" == commonpath(["../a/", "../a/b"])
    assert "../a" == commonpath(["../a/c", "../a/b"])

    assert "a/b" == commonpath(["./a/./b", "./a/b"])
    assert "a/../b" == commonpath(["./a/.././b", "a/../b/c"])


def test_commonpath_none():
    # type: () -> None

    assert "" == commonpath(["a", "b"])
    assert "" == commonpath(["bad", "b"])
    assert "" == commonpath(["a", "a", "b"])



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import subprocess
from contextlib import contextmanager
from textwrap import dedent

import pytest

from pex.atomic_directory import atomic_directory
from pex.common import safe_mkdtemp, temporary_dir
from pex.testing import PY310, ensure_python_venv, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, ContextManager, Iterator, Optional, Tuple


@pytest.fixture(scope="session")
def is_pytest_xdist(worker_id):
    # type: (str) -> bool
    return worker_id != "master"


@pytest.fixture(scope="session")
def shared_integration_test_tmpdir(
    tmpdir_factory,  # type: Any
    is_pytest_xdist,  # type: bool
):
    # type: (...) -> str
    tmpdir = str(tmpdir_factory.getbasetemp())

    # We know pytest-xdist creates a subdir under the pytest session tmp dir for each worker; so we
    # go up a level to lock a directory all workers can use.
    if is_pytest_xdist:
        tmpdir = os.path.dirname(tmpdir)

    return os.path.join(tmpdir, "shared_integration_test_tmpdir")


@pytest.fixture(scope="session")
def pex_bdist(
    pex_project_dir,  # type: str
    shared_integration_test_tmpdir,  # type: str
):
    # type: (...) -> str
    pex_bdist_chroot = os.path.join(shared_integration_test_tmpdir, "pex_bdist_chroot")
    wheels_dir = os.path.join(pex_bdist_chroot, "wheels_dir")
    with atomic_directory(pex_bdist_chroot) as chroot:
        if not chroot.is_finalized():
            pex_pex = os.path.join(chroot.work_dir, "pex.pex")
            run_pex_command(
                args=[pex_project_dir, "-o", pex_pex, "--include-tools"]
            ).assert_success()
            extract_dir = os.path.join(chroot.work_dir, "wheels_dir")
            subprocess.check_call(
                args=[pex_pex, "repository", "extract", "-f", extract_dir],
                env=make_env(PEX_TOOLS=True),
            )
    wheels = os.listdir(wheels_dir)
    assert 1 == len(wheels)
    return os.path.join(wheels_dir, wheels[0])


@pytest.fixture
def tmp_workdir():
    # type: () -> Iterator[str]
    cwd = os.getcwd()
    with temporary_dir() as tmpdir:
        os.chdir(tmpdir)
        try:
            yield os.path.realpath(tmpdir)
        finally:
            os.chdir(cwd)


@pytest.fixture(scope="module")
def mitmdump():
    # type: () -> Tuple[str, str]
    python, pip = ensure_python_venv(PY310)
    with open(os.path.join(safe_mkdtemp(), "constraints.txt"), "w") as fp:
        fp.write(
            """\
            Brotli==1.0.9
            Jinja2==2.11.3
            MarkupSafe==2.0.1
            Werkzeug==1.0.1
            asgiref==3.3.4
            blinker==1.4
            certifi==2021.10.8
            cffi==1.15.0
            click==7.1.2
            cryptography==3.2.1
            flask==1.1.4
            h11==0.13.0
            h2==4.1.0
            hpack==4.0.0
            hyperframe==6.0.1
            itsdangerous==1.1.0
            kaitaistruct==0.9
            ldap3==2.8.1
            msgpack==1.0.3
            passlib==1.7.4
            protobuf==3.13.0
            publicsuffix2==2.20191221
            pyOpenSSL==19.1.0
            pyasn1==0.4.8
            pycparser==2.21
            pyparsing==2.4.7
            pyperclip==1.8.2
            ruamel.yaml==0.16.13
            six==1.16.0
            sortedcontainers==2.2.2
            tornado==6.1
            urwid==2.1.2
            wsproto==0.15.0
            zstandard==0.14.1
            """
        )
    subprocess.check_call([pip, "install", "mitmproxy==5.3.0", "--constraint", fp.name])
    mitmdump = os.path.join(os.path.dirname(python), "mitmdump")
    return mitmdump, os.path.expanduser("~/.mitmproxy/mitmproxy-ca-cert.pem")


@pytest.fixture
def run_proxy(mitmdump, tmp_workdir):
    # type: (Tuple[str, str], str) -> Callable[[Optional[str]], ContextManager[Tuple[int, str]]]
    messages = os.path.join(tmp_workdir, "messages")
    addon = os.path.join(tmp_workdir, "addon.py")
    with open(addon, "w") as fp:
        fp.write(
            dedent(
                """\
                from mitmproxy import ctx
        
                class NotifyUp:
                    def running(self) -> None:
                        port = ctx.master.server.address[1]
                        with open({msg_channel!r}, "w") as fp:
                            print(str(port), file=fp)
        
                addons = [NotifyUp()]
                """.format(
                    msg_channel=messages
                )
            )
        )

    @contextmanager
    def _run_proxy(
        proxy_auth=None,  # type: Optional[str]
    ):
        # type: (...) -> Iterator[Tuple[int, str]]
        os.mkfifo(messages)
        proxy, ca_cert = mitmdump
        args = [proxy, "-p", "0", "-s", addon]
        if proxy_auth:
            args.extend(["--proxyauth", proxy_auth])
        proxy_process = subprocess.Popen(args)
        try:
            with open(messages, "r") as fp:
                port = int(fp.readline().strip())
                yield port, ca_cert
        finally:
            proxy_process.kill()
            os.unlink(messages)

    return _run_proxy


@pytest.fixture
def clone(tmpdir):
    # type: (Any) -> Callable[[str, str], str]

    def _clone(
        git_project_url,  # type: str
        commit,  # type: str
    ):
        project_dir = os.path.join(str(tmpdir), "project")

        subprocess.check_call(args=["git", "clone", git_project_url, project_dir])
        subprocess.check_call(
            args=["git", "config", "advice.detachedHead", "false"], cwd=project_dir
        )
        subprocess.check_call(args=["git", "checkout", commit], cwd=project_dir)
        return project_dir

    return _clone



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os


def patch():
    from pip._vendor.packaging import markers  # type: ignore[import]

    # N.B.: The following environment variable is used by the Pex runtime to control Pip and must be
    # kept in-sync with `__init__.py`.
    patched_markers_file = os.environ.pop("_PEX_PATCHED_MARKERS_FILE")
    with open(patched_markers_file) as fp:
        patched_markers = json.load(fp)

    markers.default_environment = patched_markers.copy



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys
from textwrap import dedent

import pytest

from pex.common import temporary_dir
from pex.dist_metadata import find_distribution
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraints
from pex.pep_503 import ProjectName
from pex.pex_info import PexInfo
from pex.testing import (
    PY38,
    PY39,
    PY310,
    ensure_python_interpreter,
    make_env,
    run_pex_command,
    run_simple_pex,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_interpreter_constraints_to_pex_info_py2():
    # type: () -> None
    with temporary_dir() as output_dir:
        # target python 2
        pex_out_path = os.path.join(output_dir, "pex_py2.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--interpreter-constraint=>=2.7,<3",
                "--interpreter-constraint=>=3.5",
                "-o",
                pex_out_path,
            ]
        )
        res.assert_success()
        pex_info = PexInfo.from_pex(pex_out_path)
        assert InterpreterConstraints.parse(">=2.7,<3", ">=3.5") == pex_info.interpreter_constraints


def test_interpreter_constraints_to_pex_info_py3():
    # type: () -> None
    py3_interpreter = ensure_python_interpreter(PY310)
    with temporary_dir() as output_dir:
        # target python 3
        pex_out_path = os.path.join(output_dir, "pex_py3.pex")
        res = run_pex_command(
            ["--disable-cache", "--interpreter-constraint=>3", "-o", pex_out_path],
            env=make_env(PATH=os.path.dirname(py3_interpreter)),
        )
        res.assert_success()
        pex_info = PexInfo.from_pex(pex_out_path)
        assert InterpreterConstraints.parse(">3") == pex_info.interpreter_constraints


@pytest.fixture
def satisfiable_interpreter_constraint():
    # type: () -> str
    return "=={major}.{minor}.*".format(major=sys.version_info[0], minor=sys.version_info[1])


def test_interpreter_resolution_with_constraint_option(satisfiable_interpreter_constraint):
    # type: (str) -> None
    with temporary_dir() as output_dir:
        pex_out_path = os.path.join(output_dir, "pex1.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--interpreter-constraint",
                satisfiable_interpreter_constraint,
                "-o",
                pex_out_path,
            ]
        )
        res.assert_success()
        pex_info = PexInfo.from_pex(pex_out_path)
        assert (
            InterpreterConstraints.parse(satisfiable_interpreter_constraint)
            == pex_info.interpreter_constraints
        )


def test_interpreter_resolution_with_multiple_constraint_options(
    satisfiable_interpreter_constraint,
):
    # type: (str) -> None
    with temporary_dir() as output_dir:
        pex_out_path = os.path.join(output_dir, "pex1.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--interpreter-constraint",
                satisfiable_interpreter_constraint,
                # Add a constraint that's impossible to satisfy. Because multiple
                # constraints OR, the interpreter should still resolve the 1st IC.
                "--interpreter-constraint",
                ">=500",
                "-o",
                pex_out_path,
            ]
        )
        res.assert_success()
        pex_info = PexInfo.from_pex(pex_out_path)
        assert (
            InterpreterConstraints.parse(satisfiable_interpreter_constraint, ">=500")
            == pex_info.interpreter_constraints
        )


def test_interpreter_resolution_with_pex_python_path():
    # type: () -> None

    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)

    with temporary_dir() as td:
        pexrc_path = os.path.join(td, ".pexrc")
        with open(pexrc_path, "w") as pexrc:
            pexrc.write("PEX_PYTHON_PATH={}".format(os.pathsep.join([py38, py39])))

        # constraints to build pex cleanly; PPP + pex_bootstrapper.py
        # will use these constraints to override sys.executable on pex re-exec
        interpreter_constraint = "==3.8.*" if sys.version_info[:2] == (3, 9) else "==3.9.*"

        pex_out_path = os.path.join(td, "pex.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--rcfile={}".format(pexrc_path),
                "--interpreter-constraint={}".format(interpreter_constraint),
                "-o",
                pex_out_path,
            ]
        )
        res.assert_success()

        stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload)

        assert rc == 0
        if sys.version_info[:2] == (3, 9):
            assert py38 in stdout.decode("utf-8")
        else:
            assert py39 in stdout.decode("utf-8")


def test_interpreter_constraints_honored_without_ppp_or_pp(tmpdir):
    # type: (Any) -> None
    # Create a pex with interpreter constraints, but for not the default interpreter in the path.

    py310_path = ensure_python_interpreter(PY310)
    py38_path = ensure_python_interpreter(PY38)

    pex_out_path = os.path.join(str(tmpdir), "pex.pex")
    env = make_env(
        PEX_IGNORE_RCFILES="1",
        PATH=os.pathsep.join(
            [
                os.path.dirname(py38_path),
                os.path.dirname(py310_path),
            ]
        ),
    )
    res = run_pex_command(
        ["--disable-cache", "--interpreter-constraint===%s" % PY310, "-o", pex_out_path], env=env
    )
    res.assert_success()

    # We want to try to run that pex with no environment variables set
    stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"

    stdout, rc = run_simple_pex(
        pex_out_path,
        args=["-c", "import sys; print('.'.join(map(str, sys.version_info[:2])))"],
        env=env,
    )
    assert rc == 0

    # If the constraints are honored, it will have run python3.10 and not python3.7
    # Without constraints, we would expect it to use python3.7 as it is the minimum interpreter
    # in the PATH.
    assert b"3.10\n" == stdout


def test_interpreter_resolution_pex_python_path_precedence_over_pex_python(tmpdir):
    # type: (Any) -> None

    pexrc_path = os.path.join(str(tmpdir), ".pexrc")
    ppp = os.pathsep.join(os.path.dirname(ensure_python_interpreter(py)) for py in (PY38, PY39))
    with open(pexrc_path, "w") as pexrc:
        # set both PPP and PP
        pexrc.write(
            dedent(
                """\
                PEX_PYTHON_PATH={ppp}
                PEX_PYTHON={pp}
                """.format(
                    ppp=ppp, pp=ensure_python_interpreter(PY310)
                )
            )
        )

    pex_out_path = os.path.join(str(tmpdir), "pex.pex")
    run_pex_command(
        [
            "--disable-cache",
            "--rcfile",
            pexrc_path,
            "--interpreter-constraint",
            ">=3.8,<3.10",
            "-o",
            pex_out_path,
        ]
    ).assert_success()

    print_python_version_command = [
        "-c",
        "import sys; print('.'.join(map(str, sys.version_info[:2])))",
    ]

    _, rc = run_simple_pex(pex_out_path, print_python_version_command)
    assert rc != 0, (
        "PEX_PYTHON_PATH should trump PEX_PYTHON when PEX_PYTHON is an explicit path to a Python "
        "interpreter that is not on the PEX_PYTHON_PATH and this should lead to failure to select "
        "an interpreter"
    )

    with open(pexrc_path, "w") as pexrc:
        pexrc.write(
            dedent(
                """\
                PEX_PYTHON_PATH={ppp}
                PEX_PYTHON=python
                """.format(
                    ppp=ppp
                )
            )
        )
    stdout, rc = run_simple_pex(pex_out_path, print_python_version_command)
    assert rc == 0
    assert b"3.8\n" == stdout


def test_plain_pex_exec_no_ppp_no_pp_no_constraints():
    # type: () -> None
    with temporary_dir() as td:
        pex_out_path = os.path.join(td, "pex.pex")
        env = make_env(PEX_IGNORE_RCFILES="1")
        res = run_pex_command(["--disable-cache", "-o", pex_out_path], env=env)
        res.assert_success()

        stdin_payload = b"import os, sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload, env=env)
        assert rc == 0
        assert (
            PythonInterpreter.get().resolve_base_interpreter().binary.encode() in stdout
        ), "Expected the current interpreter to be used when no constraints were supplied."


def test_pex_exec_with_pex_python_path_only():
    # type: () -> None

    py39 = ensure_python_interpreter(PY39)

    with temporary_dir() as td:
        pexrc_path = os.path.join(td, ".pexrc")
        with open(pexrc_path, "w") as pexrc:
            # set pex python path
            pexrc.write(
                "PEX_PYTHON_PATH={}".format(
                    os.pathsep.join([py39, ensure_python_interpreter(PY310)])
                )
            )

        pex_out_path = os.path.join(td, "pex.pex")
        res = run_pex_command(["--disable-cache", "--rcfile=%s" % pexrc_path, "-o", pex_out_path])
        res.assert_success()

        # test that pex bootstrapper selects the lowest version interpreter
        # in pex python path (python3.9)
        stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload)
        assert rc == 0
        assert py39 in stdout.decode("utf-8")


def test_pex_exec_with_pex_python_path_and_pex_python_but_no_constraints(tmpdir):
    # type: (Any) -> None
    pexrc_path = os.path.join(str(tmpdir), ".pexrc")
    with open(pexrc_path, "w") as pexrc:
        # set both PPP and PP
        pexrc.write(
            dedent(
                """\
                PEX_PYTHON_PATH={}
                PEX_PYTHON=python
                """.format(
                    os.pathsep.join(
                        os.path.dirname(ensure_python_interpreter(py)) for py in (PY310, PY39)
                    )
                )
            )
        )

    pex_out_path = os.path.join(str(tmpdir), "pex.pex")
    res = run_pex_command(["--disable-cache", "--rcfile", pexrc_path, "-o", pex_out_path])
    res.assert_success()

    # test that pex bootstrapper selects the lowest version interpreter
    # in pex python path (python3.9)
    stdout, rc = run_simple_pex(
        pex_out_path, args=["-c", "import sys; print('.'.join(map(str, sys.version_info[:2])))"]
    )
    assert rc == 0
    assert b"3.9\n" == stdout


def test_pex_python():
    # type: () -> None
    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    env = make_env(PATH=os.pathsep.join([os.path.dirname(py38), os.path.dirname(py39)]))
    with temporary_dir() as td:
        pexrc_path = os.path.join(td, ".pexrc")
        with open(pexrc_path, "w") as pexrc:
            pexrc.write("PEX_PYTHON={}".format(py38))

        # test PEX_PYTHON with valid constraints
        pex_out_path = os.path.join(td, "pex.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--rcfile",
                pexrc_path,
                "--interpreter-constraint",
                ">=3.8,<3.10",
                "-o",
                pex_out_path,
            ],
            env=env,
        )
        res.assert_success()

        stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload, env=env)
        assert rc == 0
        assert py38 in stdout.decode("utf-8")

        # test PEX_PYTHON with incompatible constraints
        py310 = ensure_python_interpreter(PY310)
        pexrc_path = os.path.join(td, ".pexrc")
        with open(pexrc_path, "w") as pexrc:
            pexrc.write("PEX_PYTHON={}".format(py310))

        pex_out_path = os.path.join(td, "pex2.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                "--rcfile",
                pexrc_path,
                "--interpreter-constraint",
                ">=3.8,<3.10",
                "-o",
                pex_out_path,
            ],
            env=env,
        )
        res.assert_success()

        stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload, env=env)
        assert rc == 1
        assert "Failed to find a compatible PEX_PYTHON={}.".format(py310) in stdout.decode("utf-8")

        # test PEX_PYTHON with no constraints
        pex_out_path = os.path.join(td, "pex3.pex")
        res = run_pex_command(
            ["--disable-cache", "--rcfile", pexrc_path, "-o", pex_out_path], env=env
        )
        res.assert_success()

        stdin_payload = b"import sys; print(sys.executable); sys.exit(0)"
        stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload, env=env)
        assert rc == 0
        assert py310 in stdout.decode("utf-8")


def test_interpreter_selection_using_os_environ_for_bootstrap_reexec(pex_project_dir):
    # type: (str) -> None
    """This is a test for verifying the proper function of the pex bootstrapper's interpreter
    selection logic and validate a corresponding bugfix.

    More details on the nature of the bug can be found at:
    https://github.com/pantsbuild/pex/pull/441
    """
    with temporary_dir() as td:
        pexrc_path = os.path.join(td, ".pexrc")

        # Select pexrc interpreter versions based on test environment.
        # The parent interpreter is the interpreter we expect the parent pex to
        # execute with. The child interpreter is the interpreter we expect the
        # child pex to execute with.
        if sys.version_info[:2] == (3, 10):
            child_pex_interpreter_version = PY310
        else:
            child_pex_interpreter_version = PY39

        # Write parent pex's pexrc.
        with open(pexrc_path, "w") as pexrc:
            pexrc.write("PEX_PYTHON=%s" % sys.executable)

        # The code below depends on pex.testing which depends on pytest - make sure the built pex
        # gets this dep.
        pytest_dist = find_distribution(ProjectName("pytest"))
        assert pytest_dist is not None

        test_setup_path = os.path.join(td, "setup.py")
        with open(test_setup_path, "w") as fh:
            fh.write(
                dedent(
                    """
                    from setuptools import setup

                    setup(
                        name='tester',
                        version='1.0',
                        description='tests',
                        author='tester',
                        author_email='test@test.com',
                        packages=['testing'],
                        install_requires={install_requires!r}
                    )
                    """.format(
                        install_requires=[str(pytest_dist.as_requirement())]
                    )
                )
            )

        os.mkdir(os.path.join(td, "testing"))
        test_init_path = os.path.join(td, "testing/__init__.py")
        with open(test_init_path, "w") as fh:
            fh.write(
                dedent(
                    '''
                    def tester():
                        from pex.testing import (
                            run_pex_command,
                            run_simple_pex,
                            temporary_dir
                        )
                        import os
                        from textwrap import dedent
                        with temporary_dir() as td:
                            pexrc_path = os.path.join(td, '.pexrc')
                            with open(pexrc_path, 'w') as pexrc:
                                pexrc.write("PEX_PYTHON={}")
                            test_file_path = os.path.join(td, 'build_and_run_child_pex.py')
                            with open(test_file_path, 'w') as fh:
                                fh.write(dedent("""
                                    import sys
                                    print(sys.executable)
                                """))
                            pex_out_path = os.path.join(td, 'child.pex')
                            res = run_pex_command(['--disable-cache',
                                '-o', pex_out_path])
                            stdin_payload = b'import sys; print(sys.executable); sys.exit(0)'
                            stdout, rc = run_simple_pex(pex_out_path, stdin=stdin_payload)
                            print(stdout)
                    '''.format(
                        ensure_python_interpreter(child_pex_interpreter_version)
                    )
                )
            )

        pex_out_path = os.path.join(td, "parent.pex")
        res = run_pex_command(
            [
                "--disable-cache",
                pex_project_dir,
                "{}".format(td),
                "-e",
                "testing:tester",
                "-o",
                pex_out_path,
            ]
        )
        res.assert_success()

        stdout, rc = run_simple_pex(pex_out_path)
        assert rc == 0
        # Ensure that child pex used the proper interpreter as specified by its pexrc.
        correct_interpreter_path = ensure_python_interpreter(child_pex_interpreter_version)
        assert correct_interpreter_path in stdout.decode("utf-8")



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import sys

from pex.compatibility import PY2, exec_function
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator


def iter_pth_paths(filename):
    # type: (str) -> Iterator[str]
    """Given a .pth file, extract and yield all inner paths without honoring imports.

    This shadows Python's site.py behavior, which is invoked at interpreter startup.
    """
    try:
        f = open(filename, "rU" if PY2 else "r")  # noqa
    except IOError:
        return

    seen = set()
    with f:
        for i, line in enumerate(f, start=1):
            line = line.rstrip()
            if not line or line.startswith("#"):
                continue
            elif line.startswith(("import ", "import\t")):
                # One important side effect of executing import lines can be alteration of the
                # sys.path directly or indirectly as a programmatic way to add sys.path entries
                # in contrast to the standard .pth mechanism of including fixed paths as
                # individual lines in the file. Here we capture all such programmatic attempts
                # to expand the sys.path and report the additions.
                original_sys_path = sys.path[:]
                try:
                    # N.B.: Setting sys.path to empty is ok since all the .pth files we find and
                    # execute have already been found and executed by our ambient sys.executable
                    # when it started up before running this PEX file. As such, all symbols imported
                    # by the .pth files then will still be available now as cached in sys.modules.
                    sys.path = []
                    exec_function(line, globals_map={})
                    for path in sys.path:
                        norm_path = os.path.normcase(path)
                        if norm_path not in seen:
                            yield path
                            seen.add(norm_path)
                except Exception as e:
                    # NB: import lines are routinely abused with extra code appended using `;` so
                    # the class of exceptions that might be raised in broader than ImportError. As
                    # such we catch broadly here.
                    TRACER.log(
                        "Error executing line {linenumber} of {pth_file} with content:\n"
                        "{content}\n"
                        "Error was:\n"
                        "{error}".format(linenumber=i, pth_file=filename, content=line, error=e),
                        V=9,
                    )

                    # Defer error handling to the higher level site.py logic invoked at startup.
                    return
                finally:
                    sys.path = original_sys_path
            else:
                extras_dir = os.path.abspath(os.path.join(os.path.dirname(filename), line))
                norm_extras_dir = os.path.normcase(extras_dir)
                if norm_extras_dir not in seen and os.path.exists(extras_dir):
                    yield extras_dir
                    seen.add(norm_extras_dir)
